import re

import pytest

from cognite.client.data_classes import (
    Sequence,
    SequenceData,
    SequenceDataList,
    SequenceFilter,
    SequenceList,
    SequenceUpdate,
)
from tests.utils import jsgz_load


@pytest.fixture
def mock_seq_response(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "id": 0,
                "externalId": "string",
                "name": "stringname",
                "metadata": {"metadata-key": "metadata-value"},
                "assetId": 0,
                "description": "string",
                #                "securityCategories": [0],
                "createdTime": 0,
                "lastUpdatedTime": 0,
                "columns": [
                    {
                        "externalId": "column1",
                        "valueType": "STRING",
                        "metadata": {"column-metadata-key": "column-metadata-value"},
                        "createdTime": 0,
                        "lastUpdatedTime": 0,
                    },
                    {
                        "externalId": "column2",
                        "valueType": "DOUBLE",
                        "metadata": {"column-metadata-key": "column-metadata-value"},
                        "createdTime": 0,
                        "lastUpdatedTime": 0,
                    },
                ],
            }
        ]
    }
    url_pattern = re.compile(
        re.escape(cognite_client.sequences._get_base_url_with_base_path())
        + "/sequences(?:/byids|/list|/update|/delete|/search|$|\?.+)"
    )
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_sequences_empty(rsps, cognite_client):
    response_body = {"items": []}
    url_pattern = re.compile(
        re.escape(cognite_client.sequences._get_base_url_with_base_path())
        + "/sequences(?:/byids|/update|/list|/delete|/search|$|\?.+)"
    )
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_post_sequence_data(rsps, cognite_client):
    rsps.add(
        rsps.POST, cognite_client.sequences._get_base_url_with_base_path() + "/sequences/data", status=200, json={}
    )
    yield rsps


@pytest.fixture
def mock_get_sequence_data(rsps, cognite_client):
    json = {
        "id": 0,
        "externalId": "eid",
        "columns": [{"externalId": "ceid"}],
        "rows": [{"rowNumber": 0, "values": [1]}],
    }
    rsps.add(
        rsps.POST,
        cognite_client.sequences._get_base_url_with_base_path() + "/sequences/data/list",
        status=200,
        json=json,
    )
    yield rsps


@pytest.fixture
def mock_get_sequence_empty_data(rsps, cognite_client):
    json = {"id": 0, "externalId": "eid", "columns": [{"externalId": "ceid"}, {"id": 1}], "rows": []}
    rsps.add(
        rsps.POST,
        cognite_client.sequences._get_base_url_with_base_path() + "/sequences/data/list",
        status=200,
        json=json,
    )
    yield rsps


@pytest.fixture
def mock_get_sequence_data_many_columns(rsps, cognite_client):
    json = {
        "id": 0,
        "externalId": "eid",
        "columns": [{"externalId": "ceid" + str(i)} for i in range(0, 200)],
        "rows": [{"rowNumber": 0, "values": ["str"] * 200}],
    }
    rsps.add(
        rsps.POST,
        cognite_client.sequences._get_base_url_with_base_path() + "/sequences/data/list",
        status=200,
        json=json,
    )
    yield rsps


@pytest.fixture
def mock_get_sequence_data_two_col(rsps, cognite_client):
    json = {
        "id": 0,
        "externalId": "eid",
        "columns": [{"externalId": "col1"}, {"externalId": "col2"}],
        "rows": [{"rowNumber": 0, "values": [1, 2]}],
    }
    rsps.add(
        rsps.POST,
        cognite_client.sequences._get_base_url_with_base_path() + "/sequences/data/list",
        status=200,
        json=json,
    )
    yield rsps


@pytest.fixture
def mock_get_sequence_data_two_col_with_zero(rsps, cognite_client):
    json = {
        "id": 0,
        "externalId": "eid",
        "columns": [{"externalId": "str"}, {"externalId": "lon"}],
        "rows": [{"rowNumber": 12, "values": ["string-12", 0]}],
    }
    rsps.add(
        rsps.POST,
        cognite_client.sequences._get_base_url_with_base_path() + "/sequences/data/list",
        status=200,
        json=json,
    )
    yield rsps


@pytest.fixture
def mock_get_sequence_data_with_null(rsps, cognite_client):
    json = {
        "id": 0,
        "externalId": "eid",
        "columns": [{"id": 0, "externalId": "intcol"}, {"id": 1, "externalId": "strcol"}],
        "rows": [{"rowNumber": 0, "values": [1, None]}, {"rowNumber": 1, "values": [None, "blah"]}],
    }
    rsps.add(
        rsps.POST,
        cognite_client.sequences._get_base_url_with_base_path() + "/sequences/data/list",
        status=200,
        json=json,
    )
    yield rsps


@pytest.fixture
def mock_delete_sequence_data(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.sequences._get_base_url_with_base_path() + "/sequences/data/delete",
        status=200,
        json={},
    )
    yield rsps


class TestSequences:
    def test_retrieve_single(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.retrieve(id=1)
        assert isinstance(res, Sequence)
        assert mock_seq_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_column_ids(self, cognite_client, mock_seq_response):
        seq = cognite_client.sequences.retrieve(id=1)
        assert isinstance(seq, Sequence)
        assert ["column1", "column2"] == seq.column_external_ids
        assert ["STRING", "DOUBLE"] == seq.column_value_types

    def test_retrieve_multiple(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.retrieve_multiple(ids=[1])
        assert isinstance(res, SequenceList)
        assert mock_seq_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.list()
        assert mock_seq_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list_with_filters(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.list(
            metadata={"a": "b"},
            last_updated_time={"min": 45},
            created_time={"max": 123},
            asset_ids=[1, 2],
            root_asset_ids=[1231],
            data_set_ids=[11],
            data_set_external_ids=["fml"],
            asset_subtree_ids=[1],
            asset_subtree_external_ids=["a"],
        )
        assert mock_seq_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {
            "metadata": {"a": "b"},
            "assetIds": [1, 2],
            "assetSubtreeIds": [{"id": 1}, {"externalId": "a"}],
            "rootAssetIds": [1231],
            "createdTime": {"max": 123},
            "lastUpdatedTime": {"min": 45},
            "dataSetIds": [{"id": 11}, {"externalId": "fml"}],
        } == jsgz_load(mock_seq_response.calls[0].request.body)["filter"]

    def test_create_single(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.create(Sequence(external_id="1", name="blabla", columns=[{}]))
        assert isinstance(res, Sequence)
        assert mock_seq_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)
        assert {"items": [{"name": "blabla", "externalId": "1", "columns": [{"externalId": "column0"}]}]} == jsgz_load(
            mock_seq_response.calls[0].request.body
        )

    def test_create_single_multicol(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.create(
            Sequence(
                external_id="1",
                name="blabla",
                columns=[{"valueType": "string", "last_updated_time": 123}, {"externalId": "c2"}],
            )
        )
        assert isinstance(res, Sequence)
        assert mock_seq_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)
        assert {
            "items": [
                {
                    "name": "blabla",
                    "externalId": "1",
                    "columns": [{"externalId": "column0", "valueType": "STRING"}, {"externalId": "c2"}],
                }
            ]
        } == jsgz_load(mock_seq_response.calls[0].request.body)

    def test_create_columnid_passed(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.create(
            Sequence(external_id="1", name="blabla", columns=[{"id": 1, "externalId": "a", "valueType": "STRING"}])
        )
        assert isinstance(res, Sequence)
        assert {
            "items": [{"name": "blabla", "externalId": "1", "columns": [{"valueType": "STRING", "externalId": "a"}]}]
        } == jsgz_load(mock_seq_response.calls[0].request.body)

    def test_create_column_snake_cased(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.create(
            Sequence(external_id="1", name="blabla", columns=[{"external_id": "a", "value_type": "STRING"}])
        )
        assert isinstance(res, Sequence)
        assert {
            "items": [{"name": "blabla", "externalId": "1", "columns": [{"valueType": "STRING", "externalId": "a"}]}]
        } == jsgz_load(mock_seq_response.calls[0].request.body)

    def test_create_multiple(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.create(
            [Sequence(external_id="1", name="blabla", columns=[{"externalId": "cid"}])]
        )
        assert isinstance(res, SequenceList)
        assert mock_seq_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_iter_single(self, cognite_client, mock_seq_response):
        for asset in cognite_client.sequences:
            assert mock_seq_response.calls[0].response.json()["items"][0] == asset.dump(camel_case=True)

    def test_iter_chunk(self, cognite_client, mock_seq_response):
        for assets in cognite_client.sequences(chunk_size=1):
            assert mock_seq_response.calls[0].response.json()["items"] == assets.dump(camel_case=True)

    def test_delete_single(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.delete(id=1)
        assert {"items": [{"id": 1}]} == jsgz_load(mock_seq_response.calls[0].request.body)
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.delete(id=[1])
        assert {"items": [{"id": 1}]} == jsgz_load(mock_seq_response.calls[0].request.body)
        assert res is None

    def test_update_with_resource_class(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.update(Sequence(id=1))
        assert isinstance(res, Sequence)
        assert mock_seq_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_with_update_class(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.update(SequenceUpdate(id=1).description.set("blabla"))
        assert isinstance(res, Sequence)
        assert mock_seq_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_multiple(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.update([SequenceUpdate(id=1).description.set("blabla")])
        assert isinstance(res, SequenceList)
        assert mock_seq_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_search(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.search(filter=SequenceFilter(external_id_prefix="e"))
        assert mock_seq_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {
            "search": {"name": None, "description": None, "query": None},
            "filter": {"externalIdPrefix": "e"},
            "limit": 100,
        } == jsgz_load(mock_seq_response.calls[0].request.body)

    @pytest.mark.parametrize("filter_field", ["is_string", "isString"])
    def test_search_dict_filter(self, cognite_client, mock_seq_response, filter_field):
        res = cognite_client.sequences.search(filter={filter_field: True})
        assert mock_seq_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {
            "search": {"name": None, "description": None, "query": None},
            "filter": {"isString": True},
            "limit": 100,
        } == jsgz_load(mock_seq_response.calls[0].request.body)

    def test_search_with_filter(self, cognite_client, mock_seq_response):
        res = cognite_client.sequences.search(
            name="n", description="d", query="q", filter=SequenceFilter(last_updated_time={"max": 42})
        )
        assert mock_seq_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        req_body = jsgz_load(mock_seq_response.calls[0].request.body)
        assert 42 == req_body["filter"]["lastUpdatedTime"]["max"]
        assert {"name": "n", "description": "d", "query": "q"} == req_body["search"]

    def test_update_object(self):
        assert isinstance(
            SequenceUpdate(1)
            .asset_id.set(1)
            .asset_id.set(None)
            .description.set("")
            .description.set(None)
            .external_id.set("1")
            .external_id.set(None)
            .metadata.set({})
            .name.set("")
            .name.set(None),
            SequenceUpdate,
        )

    def test_insert(self, cognite_client, mock_post_sequence_data):
        data = {i: [2 * i] for i in range(1, 11)}
        cognite_client.sequences.data.insert(column_external_ids=["0"], rows=data, external_id="eid")
        assert {
            "items": [
                {
                    "externalId": "eid",
                    "columns": ["0"],
                    "rows": [{"rowNumber": i, "values": [2 * i]} for i in range(1, 11)],
                }
            ]
        } == jsgz_load(mock_post_sequence_data.calls[0].request.body)

    def test_insert_extid(self, cognite_client, mock_post_sequence_data):
        data = {i: [2 * i] for i in range(1, 11)}
        cognite_client.sequences.data.insert(column_external_ids=["blah"], rows=data, external_id="eid")
        assert {
            "items": [
                {
                    "externalId": "eid",
                    "columns": ["blah"],
                    "rows": [{"rowNumber": i, "values": [2 * i]} for i in range(1, 11)],
                }
            ]
        } == jsgz_load(mock_post_sequence_data.calls[0].request.body)

    def test_insert_tuple(self, cognite_client, mock_post_sequence_data):
        data = [(i, [2 * i]) for i in range(1, 11)]
        cognite_client.sequences.data.insert(column_external_ids=["col0"], rows=data, external_id="eid")
        assert {
            "items": [
                {
                    "externalId": "eid",
                    "columns": ["col0"],
                    "rows": [{"rowNumber": i, "values": [2 * i]} for i in range(1, 11)],
                }
            ]
        } == jsgz_load(mock_post_sequence_data.calls[0].request.body)

    def test_insert_raw(self, cognite_client, mock_post_sequence_data):
        data = [{"rowNumber": i, "values": [2 * i, "str"]} for i in range(1, 11)]
        cognite_client.sequences.data.insert(column_external_ids=["c0"], rows=data, external_id="eid")
        assert {
            "items": [
                {
                    "externalId": "eid",
                    "columns": ["c0"],
                    "rows": [{"rowNumber": i, "values": [2 * i, "str"]} for i in range(1, 11)],
                }
            ]
        } == jsgz_load(mock_post_sequence_data.calls[0].request.body)

    def test_retrieve_no_data(self, cognite_client, mock_seq_response, mock_get_sequence_empty_data):
        data = cognite_client.sequences.data.retrieve(id=1, start=0, end=None)
        assert isinstance(data, SequenceData)
        assert 0 == len(data)
        assert 2 == len(data.column_external_ids)

    def test_retrieve_multi_data(self, cognite_client, mock_seq_response, mock_get_sequence_data_two_col_with_zero):
        data = cognite_client.sequences.data.retrieve(id=[1, 2], start=0, end=None)
        assert isinstance(data, SequenceDataList)
        assert 2 == len(data)
        assert 2 == len(data[0].column_external_ids)
        assert 2 == len(data[1].column_external_ids)

    def test_retrieve_multi_data_mixed(
        self, cognite_client, mock_seq_response, mock_get_sequence_data_two_col_with_zero
    ):
        data = cognite_client.sequences.data.retrieve(external_id=["a"], id=[1, 2], start=0, end=None)
        assert isinstance(data, SequenceDataList)
        assert 3 == len(data)
        assert 2 == len(data[0].column_external_ids)
        assert 2 == len(data[1].column_external_ids)
        assert 2 == len(data[2].column_external_ids)

    def test_retrieve_by_id(self, cognite_client, mock_seq_response, mock_get_sequence_data):
        data = cognite_client.sequences.data.retrieve(id=123, start=123, end=None)
        assert isinstance(data, SequenceData)
        assert 1 == len(data)

    def test_retrieve_by_external_id(self, cognite_client, mock_seq_response, mock_get_sequence_data):
        data = cognite_client.sequences.data.retrieve(external_id="foo", start=1, end=1000)
        assert isinstance(data, SequenceData)
        assert 1 == len(data)

    def test_retrieve_missing_column_external_id(
        self, cognite_client, mock_seq_response, mock_get_sequence_data_two_col
    ):
        data = cognite_client.sequences.data.retrieve(id=123, start=123, end=None)
        assert isinstance(data, SequenceData)
        assert 1 == len(data)

    def test_retrieve_neg_1_end_index(self, cognite_client, mock_seq_response, mock_get_sequence_data):
        cognite_client.sequences.data.retrieve(id=123, start=123, end=-1)
        assert jsgz_load(mock_get_sequence_data.calls[0].request.body)["end"] is None

    def test_delete_by_id(self, cognite_client, mock_delete_sequence_data):
        res = cognite_client.sequences.data.delete(id=1, rows=[1, 2, 3])
        assert res is None
        assert {"items": [{"id": 1, "rows": [1, 2, 3]}]} == jsgz_load(mock_delete_sequence_data.calls[0].request.body)

    def test_delete_by_external_id(self, cognite_client, mock_delete_sequence_data):
        res = cognite_client.sequences.data.delete(external_id="foo", rows=[1])
        assert res is None
        assert {"items": [{"externalId": "foo", "rows": [1]}]} == jsgz_load(
            mock_delete_sequence_data.calls[0].request.body
        )

    def test_retrieve_rows(self, cognite_client, mock_seq_response, mock_get_sequence_data):
        seq = cognite_client.sequences.retrieve(id=1)
        data = seq.rows(start=0, end=None)
        assert isinstance(data, SequenceData)
        assert 1 == len(data)

    def test_rows_helper_functions(self, cognite_client, mock_seq_response, mock_get_sequence_data):
        seq = cognite_client.sequences.retrieve(id=1)
        data = seq.rows(start=0, end=None)
        assert [1] == data[0]
        for r, v in data.items():
            assert 0 == r
            assert [1] == v
        col = data.get_column("ceid")
        assert isinstance(col, list)
        assert 1 == len(col)
        with pytest.raises(TypeError):
            data[1:23]
        with pytest.raises(ValueError):
            data.get_column("doesnotexist")

    def test_sequence_builtins(self, cognite_client, mock_seq_response):
        r1 = cognite_client.sequences.retrieve(id=0)
        r2 = cognite_client.sequences.retrieve(id=0)
        assert r1 == r2
        assert r1.__eq__(r2)
        assert str(r1) == str(r2)
        assert mock_seq_response.calls[0].response.json()["items"][0] == r1.dump(camel_case=True)

    def test_sequence_data_builtins(self, cognite_client, mock_seq_response, mock_get_sequence_data):
        r1 = cognite_client.sequences.data.retrieve(id=0, start=0, end=None)
        r2 = cognite_client.sequences.data.retrieve(id=0, start=0, end=None)
        assert r1 == r2
        assert r1.__eq__(r2)
        assert str(r1) == str(r2)
        assert r1.dump() == r2.dump()
        list_request = [call for call in mock_get_sequence_data.calls if "/data/list" in call.request.url][0]
        response = list_request.response.json()
        data_dump = r1.dump(camel_case=True)
        for f in ["columns", "rows", "id"]:
            assert response[f] == data_dump[f]


@pytest.mark.dsl
class TestSequencesPandasIntegration:
    def test_retrieve_dataframe(self, cognite_client, mock_seq_response, mock_get_sequence_data):
        df = cognite_client.sequences.data.retrieve_dataframe(external_id="foo", start=1000000, end=1100000)
        assert {"ceid"} == set(df.columns)
        assert df.shape[0] > 0
        assert df.index == [0]

    def test_retrieve_dataframe_columns_mixed(self, cognite_client, mock_seq_response, mock_get_sequence_data_two_col):
        data = cognite_client.sequences.data.retrieve(external_id="foo", start=1000000, end=1100000)
        assert isinstance(data, SequenceData)
        assert ["col1", "col2"] == list(data.to_pandas().columns)

    def test_retrieve_dataframe_multi(self, cognite_client, mock_seq_response, mock_get_sequence_data_two_col):
        data = cognite_client.sequences.data.retrieve(external_id="foo", id=[2], start=1000000, end=1100000)
        assert isinstance(data, SequenceDataList)
        assert ["col1", "col2", "col1", "col2"] == list(data.to_pandas(column_names="columnExternalId").columns)

    def test_to_pandas_invalid(self, cognite_client, mock_seq_response, mock_get_sequence_data_two_col):
        with pytest.raises(ValueError):
            cognite_client.sequences.data.retrieve_dataframe(
                external_id="foo", id=[2], start=1000000, end=1000001, column_names="id~columnExternalId"
            )
        with pytest.raises(ValueError):
            cognite_client.sequences.data.retrieve_dataframe(
                external_id="foo", id=[2], start=1000000, end=1000001, column_names="columnExternalIds"
            )

    def test_retrieve_dataframe_column_names(self, cognite_client, mock_seq_response, mock_get_sequence_data_two_col):
        data = cognite_client.sequences.data.retrieve(external_id=["eid", "eid"], start=1000000, end=1100000)
        assert isinstance(data, SequenceDataList)
        assert ["col1", "col2", "col1", "col2"] == list(data.to_pandas(column_names="columnExternalId").columns)
        assert ["eid", "eid", "eid", "eid"] == list(data.to_pandas(column_names="externalId").columns)
        assert ["eid|col1", "eid|col2", "eid|col1", "eid|col2"] == list(data.to_pandas().columns)

    def test_retrieve_dataframe_columns_mixed_with_zero(
        self, cognite_client, mock_seq_response, mock_get_sequence_data_two_col_with_zero
    ):
        import pandas as pd

        data = cognite_client.sequences.data.retrieve(external_id="foo", start=0, end=100)
        assert isinstance(data, SequenceData)
        df = data.to_pandas()
        expected_df = pd.DataFrame(index=[12], data=[["string-12", 0]], columns=["str", "lon"])
        pd.testing.assert_frame_equal(expected_df, df)

    def test_retrieve_dataframe_columns_many_extid(self, cognite_client, mock_get_sequence_data_many_columns):
        data = cognite_client.sequences.data.retrieve(external_id="foo", start=1000000, end=1100000)
        assert isinstance(data, SequenceData)
        assert ["ceid" + str(i) for i in range(200)] == list(data.to_pandas().columns)

    def test_retrieve_dataframe_convert_null(self, cognite_client, mock_seq_response, mock_get_sequence_data_with_null):
        df = cognite_client.sequences.data.retrieve_dataframe(external_id="foo", start=0, end=None)
        assert {"strcol", "intcol"} == set(df.columns)
        assert "None" not in df.strcol
        assert df.strcol.isna().any()
        assert df.intcol.isna().any()
        assert 2 == df.shape[0]

    def test_retrieve_empty_dataframe(self, cognite_client, mock_seq_response, mock_get_sequence_empty_data):
        import pandas as pd

        df = cognite_client.sequences.data.retrieve_dataframe(id=1, start=0, end=None)
        assert isinstance(df, pd.DataFrame)
        assert df.empty
        assert 2 == len(df.columns)

    def test_sequences_list_to_pandas(self, cognite_client, mock_seq_response):
        import pandas as pd

        df = cognite_client.sequences.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert 1 == df.shape[0]
        assert {"metadata-key": "metadata-value"} == df["metadata"][0]

    def test_sequences_list_to_pandas_empty(self, cognite_client, mock_sequences_empty):
        import pandas as pd

        df = cognite_client.sequences.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert df.empty

    def test_sequencess_to_pandas(self, cognite_client, mock_seq_response):
        import pandas as pd

        df = cognite_client.sequences.retrieve(id=1).to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert "metadata" not in df.columns
        assert "string" == df.loc["description"][0]
        assert "metadata-value" == df.loc["metadata-key"][0]

    def test_insert_dataframe_extids(self, cognite_client, mock_post_sequence_data):
        import pandas as pd

        df = pd.DataFrame(index=[123, 456])
        df["aa"] = [1, 2]
        df["bb"] = [5.0, 6.0]
        res = cognite_client.sequences.data.insert_dataframe(df, id=42)
        assert res is None
        request_body = jsgz_load(mock_post_sequence_data.calls[0].request.body)
        assert {
            "items": [
                {
                    "id": 42,
                    "columns": ["aa", "bb"],
                    "rows": [{"rowNumber": 123, "values": [1, 5.0]}, {"rowNumber": 456, "values": [2, 6.0]}],
                }
            ]
        } == request_body



import random
import string

import pytest

from cognite.client.data_classes import (
    OidcCredentials,
    Transformation,
    TransformationDestination,
    TransformationSchedule,
    TransformationScheduleUpdate,
)


@pytest.fixture
def new_transformation(cognite_client):
    prefix = "".join(random.choice(string.ascii_letters) for i in range(6))
    transform = Transformation(
        name="any",
        external_id=f"{prefix}-transformation",
        destination=TransformationDestination.assets(),
        query="select * from _cdf.assets",
        source_oidc_credentials=OidcCredentials(
            client_id=cognite_client.config.token_client_id,
            client_secret=cognite_client.config.token_client_secret,
            scopes=",".join(cognite_client.config.token_scopes),
            token_uri=cognite_client.config.token_url,
            cdf_project_name=cognite_client.config.project,
        ),
        destination_oidc_credentials=OidcCredentials(
            client_id=cognite_client.config.token_client_id,
            client_secret=cognite_client.config.token_client_secret,
            scopes=",".join(cognite_client.config.token_scopes),
            token_uri=cognite_client.config.token_url,
            cdf_project_name=cognite_client.config.project,
        ),
    )
    ts = cognite_client.transformations.create(transform)

    yield ts

    cognite_client.transformations.delete(id=ts.id)
    assert cognite_client.transformations.retrieve(ts.id) is None


other_transformation = new_transformation


def schedule_from_transformation(cognite_client, transformation):
    schedule = TransformationSchedule(id=transformation.id, interval="0 * * * *")
    tsc = cognite_client.transformations.schedules.create(schedule)

    yield tsc

    cognite_client.transformations.schedules.delete(id=tsc.id)
    assert cognite_client.transformations.schedules.retrieve(tsc.id) is None
    assert tsc.id == transformation.id


@pytest.fixture
def new_schedule(cognite_client, new_transformation):
    yield from schedule_from_transformation(cognite_client, new_transformation)


@pytest.fixture
def other_schedule(cognite_client, other_transformation):
    yield from schedule_from_transformation(cognite_client, other_transformation)


class TestTransformationSchedulesAPI:
    def test_create(self, new_schedule: TransformationSchedule):
        assert (
            new_schedule.interval == "0 * * * *"
            and new_schedule.is_paused == False
            and new_schedule.created_time is not None
            and new_schedule.last_updated_time is not None
        )

    def test_schedule_member(self, cognite_client, new_schedule: TransformationSchedule):
        retrieved_transformation = cognite_client.transformations.retrieve(id=new_schedule.id)
        assert (
            retrieved_transformation.schedule.interval == "0 * * * *"
            and retrieved_transformation.schedule.is_paused == False
        )

    def test_retrieve(self, cognite_client, new_schedule: TransformationSchedule):
        retrieved_schedule = cognite_client.transformations.schedules.retrieve(new_schedule.id)
        assert (
            new_schedule.id == retrieved_schedule.id
            and new_schedule.interval == retrieved_schedule.interval
            and new_schedule.is_paused == retrieved_schedule.is_paused
        )

    def test_retrieve_multiple(
        self, cognite_client, new_schedule: TransformationSchedule, other_schedule: TransformationSchedule
    ):
        assert new_schedule.id != other_schedule.id
        ids = [new_schedule.id, other_schedule.id]
        retrieved_schedules = cognite_client.transformations.schedules.retrieve_multiple(ids=ids)
        assert len(retrieved_schedules) == 2
        for retrieved_schedule in retrieved_schedules:
            assert (
                new_schedule.id == retrieved_schedule.id
                and new_schedule.interval == retrieved_schedule.interval
                and new_schedule.is_paused == retrieved_schedule.is_paused
            ) or (
                other_schedule.id == retrieved_schedule.id
                and other_schedule.interval == retrieved_schedule.interval
                and other_schedule.is_paused == retrieved_schedule.is_paused
            )

    def test_update_full(self, cognite_client, new_schedule):
        new_schedule.interval = "5 * * * *"
        new_schedule.is_paused = True
        updated_schedule = cognite_client.transformations.schedules.update(new_schedule)
        retrieved_schedule = cognite_client.transformations.schedules.retrieve(new_schedule.id)
        assert (
            updated_schedule.interval == retrieved_schedule.interval == "5 * * * *"
            and updated_schedule.is_paused == retrieved_schedule.is_paused == True
        )

    def test_update_partial(self, cognite_client, new_schedule):
        update_schedule = TransformationScheduleUpdate(id=new_schedule.id).interval.set("5 * * * *").is_paused.set(True)
        updated_schedule = cognite_client.transformations.schedules.update(update_schedule)
        retrieved_schedule = cognite_client.transformations.schedules.retrieve(new_schedule.id)
        assert (
            updated_schedule.interval == retrieved_schedule.interval == "5 * * * *"
            and updated_schedule.is_paused == retrieved_schedule.is_paused == True
        )

    def test_list(self, cognite_client, new_schedule):
        retrieved_schedules = cognite_client.transformations.schedules.list()
        assert new_schedule.id in [schedule.id for schedule in retrieved_schedules]



import re

import pytest

from cognite.client._api.raw import Database, DatabaseList, Row, RowList, Table, TableList
from cognite.client.exceptions import CogniteAPIError
from tests.utils import jsgz_load


@pytest.fixture
def mock_raw_db_response(rsps, cognite_client):
    response_body = {"items": [{"name": "db1"}]}

    url_pattern = re.compile(
        re.escape(cognite_client.raw._get_base_url_with_base_path()) + "/raw/dbs(?:/delete|$|\?.+)"
    )
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_raw_table_response(rsps, cognite_client):
    response_body = {"items": [{"name": "table1"}]}

    url_pattern = re.compile(
        re.escape(cognite_client.raw._get_base_url_with_base_path()) + "/raw/dbs/db1/tables(?:/delete|$|\?.+)"
    )
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_raw_row_response(rsps, cognite_client):
    response_body = {"items": [{"key": "row1", "columns": {"c1": 1, "c2": "2"}}]}

    raw_path_prefix = re.escape(cognite_client.raw._get_base_url_with_base_path()) + "/raw/dbs/db1/tables/table1"
    url_pattern = re.compile(raw_path_prefix + "/rows(?:/delete|/row1|$|\?.+)")
    cursors_url_pattern = re.compile(raw_path_prefix + "/cursors")
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.GET, cursors_url_pattern, status=200, json=response_body)
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_retrieve_raw_row_response(rsps, cognite_client):
    response_body = {"key": "row1", "columns": {"c1": 1, "c2": "2"}}
    rsps.add(
        rsps.GET,
        cognite_client.raw._get_base_url_with_base_path() + "/raw/dbs/db1/tables/table1/rows/row1",
        status=200,
        json=response_body,
    )
    yield rsps


class TestRawDatabases:
    def test_create_single(self, cognite_client, mock_raw_db_response):
        res = cognite_client.raw.databases.create(name="db1")
        assert isinstance(res, Database)
        assert cognite_client == res._cognite_client
        assert mock_raw_db_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)
        assert [{"name": "db1"}] == jsgz_load(mock_raw_db_response.calls[0].request.body)["items"]

    def test_create_multiple(self, cognite_client, mock_raw_db_response):
        res_list = cognite_client.raw.databases.create(name=["db1"])
        assert isinstance(res_list, DatabaseList)
        for res in res_list:
            assert cognite_client == res._cognite_client
        assert cognite_client == res_list._cognite_client
        assert [{"name": "db1"}] == jsgz_load(mock_raw_db_response.calls[0].request.body)["items"]
        assert mock_raw_db_response.calls[0].response.json()["items"] == res_list.dump(camel_case=True)

    def test_list(self, cognite_client, mock_raw_db_response):
        res_list = cognite_client.raw.databases.list()
        assert DatabaseList([Database("db1")]) == res_list

    def test_iter_single(self, cognite_client, mock_raw_db_response):
        for db in cognite_client.raw.databases:
            assert mock_raw_db_response.calls[0].response.json()["items"][0] == db.dump(camel_case=True)

    def test_iter_chunk(self, cognite_client, mock_raw_db_response):
        for db in cognite_client.raw.databases(chunk_size=1):
            assert mock_raw_db_response.calls[0].response.json()["items"] == db.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_raw_db_response):
        res = cognite_client.raw.databases.delete(name="db1")
        assert res is None
        assert [{"name": "db1"}] == jsgz_load(mock_raw_db_response.calls[0].request.body)["items"]

    def test_delete_multiple(self, cognite_client, mock_raw_db_response):
        res = cognite_client.raw.databases.delete(name=["db1"])
        assert res is None
        assert [{"name": "db1"}] == jsgz_load(mock_raw_db_response.calls[0].request.body)["items"]

    def test_delete_fail(self, cognite_client, rsps):
        rsps.add(
            rsps.POST,
            cognite_client.raw._get_base_url_with_base_path() + "/raw/dbs/delete",
            status=400,
            json={"error": {"message": "User Error", "code": 400}},
        )
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.raw.databases.delete("db1")
        assert e.value.failed == ["db1"]

    def test_get_tables_in_db(self, cognite_client, mock_raw_db_response, mock_raw_table_response):
        db = cognite_client.raw.databases.list()[0]
        tables = db.tables()
        assert TableList([Table(name="table1")]) == tables


class TestRawTables:
    def test_create_single(self, cognite_client, mock_raw_table_response):
        res = cognite_client.raw.tables.create("db1", name="table1")
        assert isinstance(res, Table)
        assert cognite_client == res._cognite_client
        assert mock_raw_table_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)
        assert [{"name": "table1"}] == jsgz_load(mock_raw_table_response.calls[0].request.body)["items"]
        assert "db1" == res._db_name

    def test_create_multiple(self, cognite_client, mock_raw_table_response):
        res_list = cognite_client.raw.tables.create("db1", name=["table1"])
        assert isinstance(res_list, TableList)
        for res in res_list:
            assert cognite_client == res._cognite_client
            assert "db1" == res._db_name
        assert cognite_client == res_list._cognite_client
        assert [{"name": "table1"}] == jsgz_load(mock_raw_table_response.calls[0].request.body)["items"]
        assert mock_raw_table_response.calls[0].response.json()["items"] == res_list.dump(camel_case=True)

    def test_list(self, cognite_client, mock_raw_table_response):
        res_list = cognite_client.raw.tables.list(db_name="db1")
        for res in res_list:
            assert "db1" == res._db_name
            assert cognite_client == res._cognite_client
        assert TableList([Table("table1")]) == res_list

    def test_iter_single(self, cognite_client, mock_raw_table_response):
        for table in cognite_client.raw.tables(db_name="db1"):
            assert mock_raw_table_response.calls[0].response.json()["items"][0] == table.dump(camel_case=True)

    def test_iter_chunk(self, cognite_client, mock_raw_table_response):
        for table_list in cognite_client.raw.tables("db1", chunk_size=1):
            for table in table_list:
                assert "db1" == table._db_name
                assert cognite_client == table._cognite_client
            assert mock_raw_table_response.calls[0].response.json()["items"] == table_list.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_raw_table_response):
        res = cognite_client.raw.tables.delete("db1", name="table1")
        assert res is None
        assert [{"name": "table1"}] == jsgz_load(mock_raw_table_response.calls[0].request.body)["items"]

    def test_delete_multiple(self, cognite_client, mock_raw_table_response):
        res = cognite_client.raw.tables.delete(db_name="db1", name=["table1"])
        assert res is None
        assert [{"name": "table1"}] == jsgz_load(mock_raw_table_response.calls[0].request.body)["items"]

    def test_delete_fail(self, cognite_client, rsps):
        rsps.add(
            rsps.POST,
            cognite_client.raw._get_base_url_with_base_path() + "/raw/dbs/db1/tables/delete",
            status=400,
            json={"error": {"message": "User Error", "code": 400}},
        )
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.raw.tables.delete("db1", "table1")
        assert e.value.failed == ["table1"]

    def test_get_rows_in_table(self, cognite_client, mock_raw_table_response, mock_raw_row_response):
        tables = cognite_client.raw.tables.list(db_name="db1")
        rows = tables[0].rows()
        assert RowList([Row._load({"key": "row1", "columns": {"c1": 1, "c2": "2"}})]) == rows


class TestRawRows:
    def test_retrieve(self, cognite_client, mock_retrieve_raw_row_response):
        res = cognite_client.raw.rows.retrieve(db_name="db1", table_name="table1", key="row1")
        assert mock_retrieve_raw_row_response.calls[0].response.json() == res.dump(camel_case=True)
        assert mock_retrieve_raw_row_response.calls[0].request.url.endswith("/rows/row1")

    def test_insert_w_rows_as_dict(self, cognite_client, mock_raw_row_response):
        res = cognite_client.raw.rows.insert(
            db_name="db1", table_name="table1", row={"row1": {"c1": 1, "c2": "2"}}, ensure_parent=True
        )
        assert res is None
        assert [{"key": "row1", "columns": {"c1": 1, "c2": "2"}}] == jsgz_load(
            mock_raw_row_response.calls[0].request.body
        )["items"]

    def test_insert_single_DTO(self, cognite_client, mock_raw_row_response):
        res = cognite_client.raw.rows.insert(
            db_name="db1", table_name="table1", row=Row(key="row1", columns={"c1": 1, "c2": "2"}), ensure_parent=False
        )
        assert res is None
        assert [{"key": "row1", "columns": {"c1": 1, "c2": "2"}}] == jsgz_load(
            mock_raw_row_response.calls[0].request.body
        )["items"]

    def test_insert_multiple_DTO(self, cognite_client, mock_raw_row_response):
        res = cognite_client.raw.rows.insert("db1", "table1", row=[Row(key="row1", columns={"c1": 1, "c2": "2"})])
        assert res is None
        assert [{"key": "row1", "columns": {"c1": 1, "c2": "2"}}] == jsgz_load(
            mock_raw_row_response.calls[0].request.body
        )["items"]

    def test_insert_fail(self, cognite_client, rsps):
        rsps.add(
            rsps.POST,
            cognite_client.raw._get_base_url_with_base_path() + "/raw/dbs/db1/tables/table1/rows",
            status=400,
            json={},
        )
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.raw.rows.insert("db1", "table1", {"row1": {"c1": 1}})
        assert e.value.failed == ["row1"]

    def test_list(self, cognite_client, mock_raw_row_response):
        res_list = cognite_client.raw.rows.list(db_name="db1", table_name="table1")
        assert RowList([Row(key="row1", columns={"c1": 1, "c2": "2"})]) == res_list
        assert "columns=" not in mock_raw_row_response.calls[0].request.path_url

    def test_list_cols(self, cognite_client, mock_raw_row_response):
        cognite_client.raw.rows.list(db_name="db1", table_name="table1", columns=["a", 1])
        assert "columns=a%2C1" in mock_raw_row_response.calls[0].request.path_url

    def test_list_cols_empty(self, cognite_client, mock_raw_row_response):
        cognite_client.raw.rows.list(db_name="db1", table_name="table1", columns=[])
        assert "columns=%2C&" in mock_raw_row_response.calls[0].request.path_url + "&"

    def test_list_cols_str_not_supported(self, cognite_client, mock_raw_row_response):
        with pytest.raises(ValueError):
            cognite_client.raw.rows.list(db_name="db1", table_name="table1", columns="a,b")

    def test_iter_single(self, cognite_client, mock_raw_row_response):
        for db in cognite_client.raw.rows(db_name="db1", table_name="table1"):
            assert mock_raw_row_response.calls[0].response.json()["items"][0] == db.dump(camel_case=True)

    def test_iter_chunk(self, cognite_client, mock_raw_row_response):
        for db in cognite_client.raw.rows("db1", "table1", chunk_size=1):
            assert mock_raw_row_response.calls[0].response.json()["items"] == db.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_raw_row_response):
        res = cognite_client.raw.rows.delete("db1", table_name="table1", key="row1")
        assert res is None
        assert [{"key": "row1"}] == jsgz_load(mock_raw_row_response.calls[0].request.body)["items"]

    def test_delete_multiple(self, cognite_client, mock_raw_row_response):
        res = cognite_client.raw.rows.delete(db_name="db1", table_name="table1", key=["row1"])
        assert res is None
        assert [{"key": "row1"}] == jsgz_load(mock_raw_row_response.calls[0].request.body)["items"]

    def test_delete_fail(self, cognite_client, rsps):
        rsps.add(
            rsps.POST,
            cognite_client.raw._get_base_url_with_base_path() + "/raw/dbs/db1/tables/table1/rows/delete",
            status=400,
            json={"error": {"message": "User Error", "code": 400}},
        )
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.raw.rows.delete("db1", "table1", "key1")
        assert e.value.failed == ["key1"]

    def test_iter(self, cognite_client, mock_raw_row_response):
        res_generator = cognite_client.raw.rows(db_name="db1", table_name="table1")
        row = next(res_generator)
        assert Row(key="row1", columns={"c1": 1, "c2": "2"}) == row
        assert "columns=" not in mock_raw_row_response.calls[0].request.path_url

    def test_iter_cols(self, cognite_client, mock_raw_row_response):
        res_generator = cognite_client.raw.rows(db_name="db1", table_name="table1", columns=["a", 1])
        next(res_generator)
        assert "columns=a%2C1" in mock_raw_row_response.calls[0].request.path_url

    def test_iter_cols_empty(self, cognite_client, mock_raw_row_response):
        res_generator = cognite_client.raw.rows(db_name="db1", table_name="table1", columns=[])
        next(res_generator)
        assert "columns=%2C&" in mock_raw_row_response.calls[0].request.path_url + "&"

    def test_iter_cols_str_not_supported(self, cognite_client, mock_raw_row_response):
        with pytest.raises(ValueError):
            cognite_client.raw.rows(db_name="db1", table_name="table1", columns="a,b")


@pytest.mark.dsl
class TestPandasIntegration:
    def test_dbs_to_pandas(self):
        import pandas as pd

        db_list = DatabaseList([Database("kar"), Database("car"), Database("dar")])

        pd.testing.assert_frame_equal(pd.DataFrame({"name": ["kar", "car", "dar"]}), db_list.to_pandas())
        pd.testing.assert_frame_equal(pd.DataFrame({"value": ["kar"]}, index=["name"]), db_list[0].to_pandas())

    def test_tables_to_pandas(self):
        import pandas as pd

        table_list = TableList([Table("kar"), Table("car"), Table("dar")])

        pd.testing.assert_frame_equal(pd.DataFrame({"name": ["kar", "car", "dar"]}), table_list.to_pandas())
        pd.testing.assert_frame_equal(pd.DataFrame({"value": ["kar"]}, index=["name"]), table_list[0].to_pandas())

    def test_rows_to_pandas(self):
        import pandas as pd

        row_list = RowList([Row("k1", {"c1": "v1", "c2": "v1"}), Row("k2", {"c1": "v2", "c2": "v2"})])
        pd.testing.assert_frame_equal(
            pd.DataFrame({"c1": ["v1", "v2"], "c2": ["v1", "v2"]}, index=["k1", "k2"]),
            row_list.to_pandas().sort_index(axis=1),
        )
        pd.testing.assert_frame_equal(pd.DataFrame({"c1": ["v1"], "c2": ["v1"]}, index=["k1"]), row_list[0].to_pandas())

    def test_rows_to_pandas_missing_cols(self):
        import pandas as pd

        row_list = RowList([Row("k1", {"c1": "v1", "c2": "v1"}), Row("k2", {"c1": "v2", "c2": "v2", "c3": "v2"})])
        pd.testing.assert_frame_equal(
            pd.DataFrame({"c1": ["v1", "v2"], "c2": ["v1", "v2"], "c3": [None, "v2"]}, index=["k1", "k2"]),
            row_list.to_pandas().sort_index(axis=1),
        )
        pd.testing.assert_frame_equal(pd.DataFrame({"c1": ["v1"], "c2": ["v1"]}, index=["k1"]), row_list[0].to_pandas())



from cognite.client.data_classes import TransformationDestination


class TestTransformationSchemaAPI:
    def test_assets(self, cognite_client):
        asset_columns = cognite_client.transformations.schema.retrieve(destination=TransformationDestination.assets())
        assert len(asset_columns) > 0
        assert len([col for col in asset_columns if col.name == "id"]) > 0
        assert [col for col in asset_columns if col.name == "id"][0].type.type == "long"
        assert [col for col in asset_columns if col.name == "metadata"][0].type.type == "map"
        assert [col for col in asset_columns if col.name == "metadata"][0].type.key_type == "string"

    def test_raw(self, cognite_client):
        asset_columns = cognite_client.transformations.schema.retrieve(destination=TransformationDestination.raw())
        assert len(asset_columns) > 0
        assert len([col for col in asset_columns if col.name == "key"]) > 0

    def test_asset_hierarchy(self, cognite_client):
        asset_columns = cognite_client.transformations.schema.retrieve(
            destination=TransformationDestination.asset_hierarchy()
        )
        assert len(asset_columns) > 0
        assert len([col for col in asset_columns if col.name == "externalId"]) > 0



import inspect
import json

import pytest

from cognite.client._api import assets, data_sets, events, files, sequences, time_series


class TestListAndIterSignatures:
    @pytest.mark.parametrize(
        "api, filter, ignore",
        [
            (
                assets.AssetsAPI,
                assets.AssetFilter,
                [
                    "root_external_ids",
                    "asset_subtree_external_ids",
                    "data_set_external_ids",
                    "aggregated_properties",
                    "partitions",
                ],
            ),
            (
                events.EventsAPI,
                events.EventFilter,
                [
                    "root_asset_external_ids",
                    "asset_subtree_external_ids",
                    "data_set_external_ids",
                    "partitions",
                    "sort",
                ],
            ),
            (
                files.FilesAPI,
                files.FileMetadataFilter,
                ["root_asset_external_ids", "data_set_external_ids", "asset_subtree_external_ids"],
            ),
            (sequences.SequencesAPI, sequences.SequenceFilter, ["asset_subtree_external_ids", "data_set_external_ids"]),
            (
                time_series.TimeSeriesAPI,
                time_series.TimeSeriesFilter,
                ["asset_subtree_external_ids", "data_set_external_ids", "include_metadata", "partitions"],
            ),
            (data_sets.DataSetsAPI, data_sets.DataSetFilter, []),
        ],
    )
    def test_list_and_iter_signatures_same_as_filter_signature(self, api, filter, ignore):
        iter_parameters = dict(inspect.signature(api.__call__).parameters)
        for name in set(ignore + ["chunk_size", "limit"]):
            del iter_parameters[name]

        list_parameters = dict(inspect.signature(api.list).parameters)
        for name in ignore + ["limit"]:
            del list_parameters[name]

        filter_parameters = dict(inspect.signature(filter.__init__).parameters)
        del filter_parameters["cognite_client"]

        iter_parameters = {v.name for _, v in iter_parameters.items()}
        filter_parameters = {v.name for _, v in filter_parameters.items()}
        list_parameters = {v.name for _, v in list_parameters.items()}

        assert iter_parameters == filter_parameters, signature_error_msg(filter_parameters, iter_parameters)
        assert list_parameters == filter_parameters, signature_error_msg(filter_parameters, list_parameters)


def signature_error_msg(expected, actual):
    pretty_expected_params = json.dumps(list(expected), indent=4, sort_keys=True)
    pretty_actual_params = json.dumps(list(actual), indent=4, sort_keys=True)
    return "Signatures don't match. \nexpected: {}\ngot: {}\n diff: {}".format(
        pretty_expected_params, pretty_actual_params, list(expected - actual) + list(actual - expected)
    )


class TestFileMetadataUploadSignatures:
    def test_upload_signatures_same_as_file_metadata_signature(self):
        upload_parameters = dict(inspect.signature(files.FilesAPI.upload).parameters)
        del upload_parameters["path"]
        del upload_parameters["recursive"]
        del upload_parameters["overwrite"]
        del upload_parameters["name"]
        del upload_parameters["security_categories"]
        upload_from_memory_parameters = dict(inspect.signature(files.FilesAPI.upload_bytes).parameters)
        del upload_from_memory_parameters["content"]
        del upload_from_memory_parameters["overwrite"]
        del upload_from_memory_parameters["name"]
        del upload_from_memory_parameters["security_categories"]
        file_metadata_parameters = dict(inspect.signature(files.FileMetadata.__init__).parameters)
        del file_metadata_parameters["id"]
        del file_metadata_parameters["uploaded_time"]
        del file_metadata_parameters["created_time"]
        del file_metadata_parameters["last_updated_time"]
        del file_metadata_parameters["uploaded"]
        del file_metadata_parameters["cognite_client"]
        del file_metadata_parameters["name"]
        del file_metadata_parameters["security_categories"]

        assert upload_parameters == file_metadata_parameters
        assert upload_from_memory_parameters == file_metadata_parameters



from cognite.client import CogniteClient
from cognite.client.exceptions import CogniteAPIError
from tests.utils import set_env_var


class TestAPIError:
    def test_api_error(self):
        e = CogniteAPIError(
            message="bla",
            code=200,
            x_request_id="abc",
            missing=[{"id": 123}],
            duplicated=[{"externalId": "abc"}],
            successful=["bla"],
        )
        assert "bla" == e.message
        assert 200 == e.code
        assert "abc" == e.x_request_id
        assert [{"id": 123}] == e.missing
        assert [{"externalId": "abc"}] == e.duplicated

        assert "bla" in e.__str__()

    def test_unknown_fields_in_api_error(self, rsps):
        with set_env_var("COGNITE_DISABLE_PYPI_VERSION_CHECK", "1"), set_env_var("COGNITE_API_KEY", "BLA"):
            c = CogniteClient()
        rsps.add(
            rsps.GET,
            c.assets._get_base_url_with_base_path() + "/any",
            status=400,
            json={"error": {"message": "bla", "extra": {"haha": "blabla"}, "other": "yup"}},
        )
        try:
            c.get(url="/api/v1/projects/{}/any".format(c.config.project))
            assert False, "Call did not raise exception"
        except CogniteAPIError as e:
            assert {"extra": {"haha": "blabla"}, "other": "yup"} == e.extra



import re
from datetime import datetime
from unittest import mock

import pytest

from cognite.client.data_classes import Datapoints, DatapointsList


@pytest.fixture(scope="session")
def test_time_series(cognite_client):
    time_series_names = ["test__constant_{}_with_noise".format(i) for i in range(0, 10)]
    time_series = {}
    for ts in cognite_client.time_series.retrieve_multiple(external_ids=time_series_names):
        value = int(re.match(r"test__constant_(\d+)_with_noise", ts.name).group(1))
        time_series[value] = ts
    yield time_series


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(
        cognite_client.datapoints.synthetic, "_post", wraps=cognite_client.datapoints.synthetic._post
    ) as _:
        yield


class TestSyntheticDatapointsAPI:
    def test_query(self, cognite_client, test_time_series, post_spy):
        query = "ts{id:%d} + ts{id:%d}" % (test_time_series[0].id, test_time_series[1].id)
        dps = cognite_client.datapoints.synthetic.query(
            expressions=query, start=datetime(2017, 1, 1), end="now", limit=23456
        )
        assert 23456 == len(dps)
        assert 3 == cognite_client.datapoints.synthetic._post.call_count

    def test_query_with_multiple_expressions(self, cognite_client, test_time_series, post_spy):
        expressions = ["ts{id:%d}" % test_time_series[0].id, "ts{id:%d}" % test_time_series[1].id]
        dps = cognite_client.datapoints.synthetic.query(
            expressions=expressions, start=datetime(2017, 1, 1), end="now", limit=23456
        )
        assert 23456 == len(dps[0])
        assert 23456 == len(dps[1])
        assert 6 == cognite_client.datapoints.synthetic._post.call_count

    def test_query_with_errors(self, cognite_client, test_time_series, post_spy):
        dps = cognite_client.datapoints.synthetic.query(
            expressions=["A / (B - B)"],
            start=datetime(2017, 1, 1),
            end="now",
            limit=100,
            variables={"A": test_time_series[0], "B": test_time_series[1]},
        )[0]
        assert 100 == len(dps)
        assert 100 == len(dps.error)
        assert all(x is not None for x in dps.error)
        assert all(x is None for x in dps.value)
        assert (100, 1) == dps.to_pandas().shape
        assert (100, 2) == dps.to_pandas(include_errors=True).shape

    @pytest.mark.dsl
    def test_expression_builder_time_series_vs_string(self, cognite_client, test_time_series):
        from sympy import symbols

        dps1 = cognite_client.datapoints.synthetic.query(
            expressions=symbols("a"),
            start=datetime(2017, 1, 1),
            end="now",
            limit=100,
            variables={"a": test_time_series[0].external_id},
        )
        dps2 = cognite_client.datapoints.synthetic.query(
            expressions=[symbols("a"), symbols("b")],
            start=datetime(2017, 1, 1),
            end="now",
            limit=100,
            variables={"a": test_time_series[0], "b": test_time_series[0].external_id},
        )
        assert 100 == len(dps1)
        assert 100 == len(dps2[0])
        assert dps1 == dps2[0]
        assert isinstance(dps1, Datapoints)
        assert isinstance(dps2, DatapointsList)

    @pytest.mark.dsl
    def test_expression_builder_complex(self, cognite_client, test_time_series):
        from sympy import cos, log, pi, sin, sqrt, symbols

        abc = list("abcdefghij")
        syms = symbols(abc)
        expression = syms[0]
        for s in syms:
            expression = expression + s
        expression = (
            (expression * expression)
            + sqrt(sin(pi * 0.1 ** syms[1]))
            + log(23 + syms[5] ** 1.234)
            + cos(syms[3] ** (1 + 0.1 ** syms[4]))
            + sqrt(log(abs(syms[8]) + 1))
        )
        dps1 = cognite_client.datapoints.synthetic.query(
            expressions=[expression],
            start=datetime(2017, 1, 1),
            end="now",
            limit=100,
            variables={v: test_time_series[tsi] for v, tsi in zip(abc, range(10))},
        )[0]
        assert 100 == len(dps1)



import uuid

import pytest

from cognite.client.data_classes import (
    ConstantResolver,
    TemplateGroup,
    TemplateGroupList,
    TemplateGroupVersion,
    TemplateGroupVersionList,
    TemplateInstance,
    TemplateInstanceList,
)
from cognite.client.data_classes.events import Event
from cognite.client.data_classes.templates import Source, TemplateInstanceUpdate, View, ViewResolver
from cognite.client.exceptions import CogniteNotFoundError


@pytest.fixture
def new_template_group(cognite_client):
    external_id = uuid.uuid4().hex[0:20]
    username = cognite_client.iam.token.inspect().subject
    template_group = cognite_client.templates.groups.create(
        TemplateGroup(
            external_id=external_id, description="some description", owners=[username, external_id + "@cognite.com"]
        )
    )
    yield template_group, external_id
    cognite_client.templates.groups.delete(external_ids=external_id)
    assert cognite_client.templates.groups.retrieve_multiple(external_ids=template_group.external_id) is None


@pytest.fixture
def new_template_group_version(cognite_client, new_template_group):
    new_group, ext_id = new_template_group
    schema = """
    type Demographics @template {
        "The amount of people"
        populationSize: Int,
        "The population growth rate"
        growthRate: Float,
    }

    type Country @template {
        name: String,
        testView: Long,
        demographics: Demographics,
        deaths: TimeSeries,
        confirmed: TimeSeries,
    }"""
    version = TemplateGroupVersion(schema)
    new_version = cognite_client.templates.versions.upsert(ext_id, version=version)
    yield new_group, ext_id, new_version
    print(ext_id, new_version.version)
    cognite_client.templates.versions.delete(ext_id, new_version.version)


@pytest.fixture
def new_template_instance(cognite_client, new_template_group_version):
    new_group, ext_id, new_version = new_template_group_version
    template_instance_1 = TemplateInstance(
        external_id="norway",
        template_name="Country",
        field_resolvers={
            "name": ConstantResolver("Norway"),
            "testView": ViewResolver("test", {"type": "bar"}),
            "demographics": ConstantResolver("norway_demographics"),
            "deaths": ConstantResolver("Norway_deaths"),
            "confirmed": ConstantResolver("Norway_confirmed"),
        },
    )
    instance = cognite_client.templates.instances.create(ext_id, new_version.version, template_instance_1)
    yield new_group, ext_id, new_version, instance
    cognite_client.templates.instances.delete(ext_id, new_version.version, instance.external_id)


@pytest.fixture
def new_view(cognite_client, new_template_group_version):
    events = []
    for i in range(0, 1001):
        events.append(Event(external_id="test_evt_templates_1_" + str(i), type="test_templates_1", start_time=i * 1000))
    try:
        cognite_client.events.create(events)
    except:
        # We only generate this data once for a given project, to prevent issues with eventual consistency etc.
        None

    new_group, ext_id, new_version = new_template_group_version
    view = View(
        external_id="test",
        source=Source(
            type="events",
            filter={"startTime": {"min": "$minStartTime"}, "type": "test_templates_1"},
            mappings={"test_type": "type", "startTime": "startTime"},
        ),
    )
    view = cognite_client.templates.views.create(ext_id, new_version.version, view)
    yield new_group, ext_id, new_version, view
    try:
        cognite_client.templates.views.delete(ext_id, new_version.version, view.external_id)
    except:
        None


class TestTemplatescognite_client:
    def test_groups_get_single(self, cognite_client, new_template_group):
        new_group, ext_id = new_template_group
        res = cognite_client.templates.groups.retrieve_multiple(external_ids=[new_group.external_id])
        assert isinstance(res[0], TemplateGroup)
        assert new_group.external_id == ext_id

    def test_groups_retrieve_unknown(self, cognite_client, new_template_group):
        with pytest.raises(CogniteNotFoundError):
            cognite_client.templates.groups.retrieve_multiple(external_ids=["this does not exist"])
        assert cognite_client.templates.groups.retrieve_multiple(external_ids="this does not exist") is None

    def test_groups_list_filter(self, cognite_client, new_template_group):
        new_group, ext_id = new_template_group
        res = cognite_client.templates.groups.list(owners=[ext_id + "@cognite.com"])
        assert len(res) == 1
        assert isinstance(res, TemplateGroupList)

    def test_groups_upsert(self, cognite_client, new_template_group):
        new_group, ext_id = new_template_group
        res = cognite_client.templates.groups.upsert(TemplateGroup(ext_id))
        assert isinstance(res, TemplateGroup)

    def test_versions_list(self, cognite_client, new_template_group_version):
        new_group, ext_id, new_version = new_template_group_version
        res = cognite_client.templates.versions.list(ext_id)
        assert len(res) == 1
        assert isinstance(res, TemplateGroupVersionList)

    def test_instances_get_single(self, cognite_client, new_template_instance):
        new_group, ext_id, new_version, new_instance = new_template_instance
        res = cognite_client.templates.instances.retrieve_multiple(
            ext_id, new_version.version, new_instance.external_id
        )
        assert isinstance(res, TemplateInstance)
        assert res.external_id == new_instance.external_id

    def test_instances_list(self, cognite_client, new_template_instance):
        new_group, ext_id, new_version, new_instance = new_template_instance
        res = cognite_client.templates.instances.list(ext_id, new_version.version, template_names=["Country"])
        assert isinstance(res, TemplateInstanceList)
        assert len(res) == 1

    def test_instances_upsert(self, cognite_client, new_template_instance):
        new_group, ext_id, new_version, new_instance = new_template_instance
        upserted_instance = TemplateInstance(
            external_id="norway",
            template_name="Demographics",
            field_resolvers={"populationSize": ConstantResolver(5328000), "growthRate": ConstantResolver(value=0.02)},
        )
        res = cognite_client.templates.instances.upsert(ext_id, new_version.version, upserted_instance)
        assert res.external_id == new_instance.external_id

    def test_instances_update_add(self, cognite_client, new_template_instance):
        new_group, ext_id, new_version, new_instance = new_template_instance
        upserted_instance = TemplateInstanceUpdate(external_id="norway").field_resolvers.add(
            {"name": ConstantResolver("Patched")}
        )
        res = cognite_client.templates.instances.update(ext_id, new_version.version, upserted_instance)
        assert res.external_id == new_instance.external_id and res.field_resolvers["name"] == ConstantResolver(
            "Patched"
        )

    def test_instances_update_remove(self, cognite_client, new_template_instance):
        new_group, ext_id, new_version, new_instance = new_template_instance
        upserted_instance = TemplateInstanceUpdate(external_id="norway").field_resolvers.remove(["name"])
        res = cognite_client.templates.instances.update(ext_id, new_version.version, upserted_instance)
        assert res.external_id == new_instance.external_id and "name" not in res.field_resolvers

    def test_instances_update_set(self, cognite_client, new_template_instance):
        new_group, ext_id, new_version, new_instance = new_template_instance
        upserted_instance = TemplateInstanceUpdate(external_id="norway").field_resolvers.set(
            {"name": ConstantResolver("Patched")}
        )
        res = cognite_client.templates.instances.update(ext_id, new_version.version, upserted_instance)
        assert res.external_id == new_instance.external_id and res.field_resolvers == {
            "name": ConstantResolver("Patched")
        }

    def test_query(self, cognite_client, new_template_instance):
        new_group, ext_id, new_version, new_instance = new_template_instance
        query = """
        { 
        countryList 
            { 
                name, 
                deaths { 
                    externalId, 
                    datapoints(limit: 2) { 
                        timestamp, value 
                    } 
                }, 
                confirmed { 
                    externalId, 
                    datapoints(limit: 2) { 
                        timestamp, value 
                    } 
                } 
            } 
        }
        """
        res = cognite_client.templates.graphql_query(ext_id, 1, query)
        assert res.data is not None

    def test_view_list(self, cognite_client, new_view):
        new_group, ext_id, new_version, view = new_view
        first_element = [
            res
            for res in cognite_client.templates.views.list(ext_id, new_version.version)
            if res.external_id == view.external_id
        ][0]
        assert first_element == view

    def test_view_delete(self, cognite_client, new_view):
        new_group, ext_id, new_version, view = new_view
        cognite_client.templates.views.delete(ext_id, new_version.version, [view.external_id])
        assert (
            len(
                [
                    res
                    for res in cognite_client.templates.views.list(ext_id, new_version.version)
                    if res.external_id == view.external_id
                ]
            )
            == 0
        )

    def test_view_resolve(self, cognite_client, new_view):
        new_group, ext_id, new_version, view = new_view
        res = cognite_client.templates.views.resolve(
            ext_id, new_version.version, view.external_id, input={"minStartTime": 10 * 1000}, limit=10
        )
        assert res == [{"startTime": (i + 10) * 1000, "test_type": "test_templates_1"} for i in range(0, 10)]

    def test_view_resolve_pagination(self, cognite_client, new_view):
        new_group, ext_id, new_version, view = new_view
        res = cognite_client.templates.views.resolve(
            ext_id, new_version.version, view.external_id, input={"minStartTime": 0}, limit=-1
        )
        assert res == [{"startTime": i * 1000, "test_type": "test_templates_1"} for i in range(0, 1001)]

    def test_view_upsert(self, cognite_client, new_view):
        new_group, ext_id, new_version, view = new_view
        view.source.mappings["another_type"] = "type"
        res = cognite_client.templates.views.upsert(ext_id, new_version.version, [view])
        assert res == view



import json
import math
from datetime import datetime
from random import random
from typing import List
from unittest import mock

import pytest

from cognite.client import utils
from cognite.client._api.datapoints import DatapointsBin, DatapointsFetcher, _DPTask, _DPWindow
from cognite.client.data_classes import Datapoint, Datapoints, DatapointsList, DatapointsQuery
from cognite.client.exceptions import CogniteAPIError, CogniteDuplicateColumnsError, CogniteNotFoundError
from tests.utils import jsgz_load, set_request_limit


def generate_datapoints(start: int, end: int, aggregates=None, granularity=None):
    dps = []
    granularity = utils._time.granularity_to_ms(granularity) if granularity else 1000
    for i in range(start, end, granularity):
        dp = {}
        if aggregates:
            if aggregates == ["count"]:
                dp["count"] = int(math.ceil((end - start) / 1000))
            else:
                for agg in aggregates:
                    dp[agg] = random()
        else:
            dp["value"] = random()
        dp["timestamp"] = i
        dps.append(dp)
    return dps


@pytest.fixture
def mock_get_datapoints(rsps, cognite_client):
    def request_callback(request):
        payload = jsgz_load(request.body)

        items = []
        for dps_query in payload["items"]:
            aggregates = []

            if "aggregates" in dps_query:
                aggregates = dps_query["aggregates"]
            elif "aggregates" in payload:
                aggregates = payload["aggregates"]

            granularity = None
            if "granularity" in dps_query:
                granularity = dps_query["granularity"]
            elif "granularity" in payload:
                granularity = payload["granularity"]

            if (granularity and not aggregates) or (not granularity and aggregates):
                return (
                    400,
                    {},
                    json.dumps({"error": {"code": 400, "message": "You must specify both aggregates AND granularity"}}),
                )

            if "start" in dps_query and "end" in dps_query:
                start, end = dps_query["start"], dps_query["end"]
            else:
                start, end = payload["start"], payload["end"]

            limit = 100000
            if "limit" in dps_query:
                limit = dps_query["limit"]
            elif "limit" in payload:
                limit = payload["limit"]

            dps = generate_datapoints(start, end, aggregates, granularity)
            dps = dps[:limit]
            id_to_return = dps_query.get("id", int(dps_query.get("externalId", "-1")))
            external_id_to_return = dps_query.get("externalId", str(dps_query.get("id", -1)))
            items.append(
                {
                    "id": id_to_return,
                    "externalId": external_id_to_return,
                    "isString": False,
                    "isStep": False,
                    "datapoints": dps,
                }
            )
        response = {"items": items}
        return 200, {}, json.dumps(response)

    rsps.add_callback(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        callback=request_callback,
        content_type="application/json",
    )
    yield rsps


@pytest.fixture
def mock_get_datapoints_empty(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        status=200,
        json={
            "items": [{"id": 1, "externalId": "1", "isString": False, "isStep": False, "unit": "kPa", "datapoints": []}]
        },
    )
    yield rsps


@pytest.fixture
def mock_get_datapoints_include_outside(rsps, cognite_client):
    # return 100001 datapoints with one beyond 'end'
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        status=200,
        json={
            "items": [
                {
                    "id": 1,
                    "externalId": "1",
                    "isString": False,
                    "isStep": False,
                    "datapoints": [{"timestamp": i, "value": i} for i in range(1000000000, 1000000000 + 100001)],
                }
            ]
        },
    )
    yield rsps


@pytest.fixture
def mock_get_datapoints_one_ts_empty(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        status=200,
        json={
            "items": [
                {
                    "id": 1,
                    "externalId": "1",
                    "isString": False,
                    "isStep": False,
                    "datapoints": [{"timestamp": 1, "value": 1}],
                }
            ]
        },
    )
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        status=200,
        json={"items": [{"id": 2, "externalId": "2", "isString": False, "isStep": False, "datapoints": []}]},
    )
    yield rsps


@pytest.fixture
def mock_get_datapoints_one_ts_has_missing_aggregates(rsps, cognite_client):
    def callback(request):
        item = jsgz_load(request.body)
        if item["aggregates"] == ["average"]:
            dps = {
                "id": 1,
                "externalId": "abc",
                "isString": False,
                "isStep": False,
                "datapoints": [
                    {"timestamp": 0, "average": 0},
                    {"timestamp": 1, "average": 1},
                    {"timestamp": 2, "average": 2},
                    {"timestamp": 3, "average": 3},
                    {"timestamp": 4, "average": 4},
                ],
            }
        else:
            dps = {
                "id": 2,
                "externalId": "def",
                "isString": False,
                "isStep": False,
                "datapoints": [
                    {"timestamp": 0},
                    {"timestamp": 1, "interpolation": 1},
                    {"timestamp": 2},
                    {"timestamp": 3, "interpolation": 3},
                    {"timestamp": 4},
                ],
            }
        return 200, {}, json.dumps({"items": [dps]})

    rsps.add_callback(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        callback=callback,
        content_type="application/json",
    )
    yield rsps


@pytest.fixture
def mock_get_datapoints_several_missing(rsps, cognite_client):
    def callback(request):
        item = jsgz_load(request.body)
        if item["aggregates"] == ["interpolation"]:
            dps = {
                "id": 2,
                "externalId": "abc",
                "isString": False,
                "isStep": False,
                "datapoints": [{"timestamp": 1000, "interpolation": 1}, {"timestamp": 3000, "interpolation": 3}],
            }
        elif item["aggregates"] == ["count"]:
            dps = {
                "id": 3,
                "externalId": "def",
                "isString": False,
                "isStep": False,
                "datapoints": [
                    {"timestamp": 1000, "count": 2},
                    {"timestamp": 3000, "count": 4},
                    {"timestamp": 4000, "count": 5},
                ],
            }
        elif item["aggregates"] == ["average"]:
            dps = {
                "id": 1,
                "externalId": "def",
                "isString": False,
                "isStep": False,
                "datapoints": [
                    {"timestamp": 0, "average": 11},
                    {"timestamp": 1000, "average": 22},
                    {"timestamp": 3000, "average": 44},
                ],
            }

        return 200, {}, json.dumps({"items": [dps]})

    rsps.add_callback(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        callback=callback,
        content_type="application/json",
    )

    response_body = {
        "items": [
            {"id": 1, "isStep": False, "isString": False},
            {"id": 2, "isStep": None, "isString": False},
            {"id": 3, "isStep": True, "isString": False},
        ]
    }
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/byids",
        status=200,
        json=response_body,
    )
    rsps.assert_all_requests_are_fired = False
    yield rsps


@pytest.fixture
def mock_get_datapoints_single_isstep(rsps, cognite_client):
    def callback(request):
        item = jsgz_load(request.body)
        if item["aggregates"] == ["interpolation"]:
            dps = {
                "id": 3,
                "externalId": "abc",
                "isStep": True,
                "isString": False,
                "datapoints": [{"timestamp": 1000, "interpolation": 1}, {"timestamp": 3000, "interpolation": 3}],
            }
        return 200, {}, json.dumps({"items": [dps]})

    rsps.add_callback(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        callback=callback,
        content_type="application/json",
    )
    yield rsps


@pytest.fixture
def set_dps_workers(cognite_client):
    def set_workers(limit):
        cognite_client.datapoints._config.max_workers = limit

    workers_tmp = cognite_client.datapoints._config.max_workers
    yield set_workers
    cognite_client.datapoints._config.max_workers = workers_tmp


def assert_dps_response_is_correct(calls, dps_object):
    datapoints = []
    for call in calls:
        if jsgz_load(call.request.body)["limit"] > 1 and jsgz_load(call.request.body).get("aggregates") != ["count"]:
            dps_response = call.response.json()["items"][0]
            if dps_response["id"] == dps_object.id and dps_response["externalId"] == dps_object.external_id:
                datapoints.extend(dps_response["datapoints"])
                id = dps_response["id"]
                external_id = dps_response["externalId"]

    expected_dps = sorted(datapoints, key=lambda x: x["timestamp"])
    assert id == dps_object.id
    assert external_id == dps_object.external_id
    assert expected_dps == dps_object.dump(camel_case=True)["datapoints"]


class TestGetDatapoints:
    def test_retrieve_datapoints_by_id(self, cognite_client, mock_get_datapoints):
        dps_res = cognite_client.datapoints.retrieve(id=123, start=1000000, end=1100000)
        assert isinstance(dps_res, Datapoints)
        assert_dps_response_is_correct(mock_get_datapoints.calls, dps_res)

    def test_retrieve_datapoints_500(self, cognite_client, rsps):
        rsps.add(
            rsps.POST,
            cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
            json={"error": {"code": 500, "message": "Internal Server Error"}},
            status=500,
        )
        with pytest.raises(CogniteAPIError):
            cognite_client.datapoints.retrieve(id=123, start=1000000, end=1100000)

    def test_retrieve_datapoints_by_external_id(self, cognite_client, mock_get_datapoints):
        dps_res = cognite_client.datapoints.retrieve(external_id="123", start=1000000, end=1100000)
        assert_dps_response_is_correct(mock_get_datapoints.calls, dps_res)

    def test_retrieve_datapoints_aggregates(self, cognite_client, mock_get_datapoints):
        dps_res = cognite_client.datapoints.retrieve(
            id=123, start=1000000, end=1100000, aggregates=["average", "stepInterpolation"], granularity="10s"
        )
        assert_dps_response_is_correct(mock_get_datapoints.calls, dps_res)

    def test_retrieve_datapoints_local_aggregates(self, cognite_client, mock_get_datapoints):
        dps_res_list = cognite_client.datapoints.retrieve(
            external_id={"externalId": "123", "aggregates": ["average"]},
            id={"id": 234},
            start=1000000,
            end=1100000,
            aggregates=["max"],
            granularity="10s",
        )
        for dps_res in dps_res_list:
            assert_dps_response_is_correct(mock_get_datapoints.calls, dps_res)

    def test_retrieve_datapoints_some_aggregates_omitted(
        self, cognite_client, mock_get_datapoints_one_ts_has_missing_aggregates
    ):
        dps_res_list = cognite_client.datapoints.retrieve(
            id={"id": 1, "aggregates": ["average"]},
            external_id={"externalId": "def", "aggregates": ["interpolation"]},
            start=0,
            end=1,
            granularity="1s",
        )
        for dps in dps_res_list:
            if dps.id == 1:
                assert dps.average == [0, 1, 2, 3, 4]
            elif dps.id == 2:
                assert dps.interpolation == [None, 1, None, 3, None]

    def test_datapoints_paging(self, cognite_client, mock_get_datapoints, set_dps_workers):
        set_dps_workers(1)
        with set_request_limit(cognite_client.datapoints, 2):
            dps_res = cognite_client.datapoints.retrieve(
                id=123, start=0, end=10000, aggregates=["average"], granularity="1s"
            )
        assert 6 == len(mock_get_datapoints.calls)
        assert 10 == len(dps_res)

    def test_datapoints_concurrent(self, cognite_client, mock_get_datapoints):
        with set_request_limit(cognite_client.datapoints, 20):
            dps_res = cognite_client.datapoints.retrieve(
                id=123, start=0, end=100000, aggregates=["average"], granularity="1s"
            )
        requested_windows = sorted(
            [
                (jsgz_load(call.request.body)["start"], jsgz_load(call.request.body)["end"])
                for call in mock_get_datapoints.calls
            ],
            key=lambda x: x[0],
        )
        assert (0, 100000) == requested_windows[0]
        assert [(20000, 100000), (40000, 100000), (60000, 100000), (80000, 100000)] == requested_windows[2:]
        assert_dps_response_is_correct(mock_get_datapoints.calls, dps_res)

    def test_datapoints_paging_with_limit(self, cognite_client, mock_get_datapoints):
        with set_request_limit(cognite_client.datapoints, 3):
            dps_res = cognite_client.datapoints.retrieve(
                id=123, start=0, end=10000, aggregates=["average"], granularity="1s", limit=4
            )
        assert 4 == len(dps_res)

    def test_retrieve_datapoints_multiple_time_series(self, cognite_client, mock_get_datapoints):
        ids = [1, 2, 3]
        external_ids = ["4", "5", "6"]
        dps_res_list = cognite_client.datapoints.retrieve(id=ids, external_id=external_ids, start=0, end=100000)
        assert isinstance(dps_res_list, DatapointsList), type(dps_res_list)
        for dps_res in dps_res_list:
            if dps_res.id in ids:
                ids.remove(dps_res.id)
            if dps_res.external_id in external_ids:
                external_ids.remove(dps_res.external_id)
            assert_dps_response_is_correct(mock_get_datapoints.calls, dps_res)
        assert 0 == len(ids)
        assert 0 == len(external_ids)

    def test_retrieve_datapoints_empty(self, cognite_client, mock_get_datapoints_empty):
        res = cognite_client.datapoints.retrieve(id=1, start=0, end=10000)
        assert 0 == len(res)

    def test_retrieve_datapoints_empty_extrafields_set(self, cognite_client, mock_get_datapoints_empty):
        res = cognite_client.datapoints.retrieve(id=1, start=0, end=10000)
        assert "kPa" == res.unit
        assert res.is_step is False
        assert res.is_string is False

    def test_aggregate_limits_correct(self, cognite_client, mock_get_datapoints):
        cognite_client.datapoints.retrieve(id={"id": 1, "aggregates": ["average"]}, start=0, end=10, granularity="1d")
        cognite_client.datapoints.retrieve(id=1, start=0, end=10, granularity="1d", aggregates=["max"])
        cognite_client.datapoints.retrieve(id=1, start=0, end=10)
        assert 10000 == jsgz_load(mock_get_datapoints.calls[0].request.body)["limit"]
        assert 10000 == jsgz_load(mock_get_datapoints.calls[1].request.body)["limit"]
        assert 100000 == jsgz_load(mock_get_datapoints.calls[2].request.body)["limit"]


class TestQueryDatapoints:
    def test_query_single(self, cognite_client, mock_get_datapoints):
        dps_res = cognite_client.datapoints.query(query=DatapointsQuery(id=1, start=0, end=10000))
        assert isinstance(dps_res, DatapointsList)
        assert_dps_response_is_correct(mock_get_datapoints.calls, dps_res[0])

    def test_query_multiple(self, cognite_client, mock_get_datapoints):
        dps_res_list = cognite_client.datapoints.query(
            query=[
                DatapointsQuery(id=1, start=0, end=10000),
                DatapointsQuery(external_id="2", start=10000, end=20000, aggregates=["average"], granularity="2s"),
            ]
        )
        assert isinstance(dps_res_list, List)
        for dps_res in dps_res_list:
            assert_dps_response_is_correct(mock_get_datapoints.calls, dps_res[0])

    def test_query_empty(self, cognite_client, mock_get_datapoints_empty):
        dps_res = cognite_client.datapoints.query(query=DatapointsQuery(id=1, start=0, end=10000))
        assert isinstance(dps_res, DatapointsList)
        assert 1 == len(dps_res)
        assert 0 == len(dps_res[0])


@pytest.fixture
def mock_retrieve_latest(rsps, cognite_client):
    def request_callback(request):
        payload = jsgz_load(request.body)

        items = []
        for latest_query in payload["items"]:
            id = latest_query.get("id", -1)
            external_id = latest_query.get("externalId", "-1")
            before = latest_query.get("before", 10001)
            items.append(
                {
                    "id": id,
                    "externalId": external_id,
                    "isString": False,
                    "isStep": False,
                    "datapoints": [{"timestamp": before - 1, "value": random()}],
                }
            )
        return 200, {}, json.dumps({"items": items})

    rsps.add_callback(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/latest",
        callback=request_callback,
        content_type="application/json",
    )
    yield rsps


@pytest.fixture
def mock_retrieve_latest_empty(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/latest",
        status=200,
        json={
            "items": [
                {"id": 1, "externalId": "1", "isString": False, "isStep": True, "datapoints": []},
                {"id": 2, "isString": False, "isStep": False, "externalId": "2", "datapoints": []},
            ]
        },
    )
    yield rsps


@pytest.fixture
def mock_retrieve_latest_with_failure(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/latest",
        status=200,
        json={
            "items": [
                {"id": 1, "externalId": "1", "isString": False, "isStep": False, "datapoints": []},
                {"id": 2, "externalId": "2", "isString": False, "isStep": False, "datapoints": []},
            ]
        },
    )
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/latest",
        status=500,
        json={"error": {"code": 500, "message": "Internal Server Error"}},
    )
    yield rsps


class TestGetLatest:
    def test_retrieve_latest(self, cognite_client, mock_retrieve_latest):
        res = cognite_client.datapoints.retrieve_latest(id=1)
        assert isinstance(res, Datapoints)
        assert 10000 == res[0].timestamp
        assert isinstance(res[0].value, float)

    def test_retrieve_latest_multiple_ts(self, cognite_client, mock_retrieve_latest):
        res = cognite_client.datapoints.retrieve_latest(id=1, external_id="2")
        assert isinstance(res, DatapointsList)
        for dps in res:
            assert 10000 == dps[0].timestamp
            assert isinstance(dps[0].value, float)

    def test_retrieve_latest_with_before(self, cognite_client, mock_retrieve_latest):
        res = cognite_client.datapoints.retrieve_latest(id=1, before=10)
        assert isinstance(res, Datapoints)
        assert 9 == res[0].timestamp
        assert isinstance(res[0].value, float)

    def test_retrieve_latest_multiple_ts_with_before(self, cognite_client, mock_retrieve_latest):
        res = cognite_client.datapoints.retrieve_latest(id=[1, 2], external_id=["1", "2"], before=10)
        assert isinstance(res, DatapointsList)
        for dps in res:
            assert 9 == dps[0].timestamp
            assert isinstance(dps[0].value, float)

    def test_retrieve_latest_empty(self, cognite_client, mock_retrieve_latest_empty):
        res = cognite_client.datapoints.retrieve_latest(id=1)
        assert isinstance(res, Datapoints)
        assert 0 == len(res)

    def test_retrieve_latest_multiple_ts_empty(self, cognite_client, mock_retrieve_latest_empty):
        res_list = cognite_client.datapoints.retrieve_latest(id=[1, 2])
        assert isinstance(res_list, DatapointsList)
        assert 2 == len(res_list)
        for res in res_list:
            assert 0 == len(res)

    def test_retrieve_latest_concurrent_fails(self, cognite_client, mock_retrieve_latest_with_failure):
        with set_request_limit(cognite_client.datapoints, 2):
            with pytest.raises(CogniteAPIError) as e:
                cognite_client.datapoints.retrieve_latest(id=[1, 2, 3])
            assert e.value.code == 500


@pytest.fixture
def mock_post_datapoints(rsps, cognite_client):
    rsps.add(
        rsps.POST, cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data", status=200, json={}
    )
    yield rsps


@pytest.fixture
def mock_post_datapoints_400(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data",
        status=400,
        json={"error": {"message": "Ts not found", "missing": [{"externalId": "does_not_exist"}]}},
    )
    yield rsps


class TestInsertDatapoints:
    def test_insert_tuples(self, cognite_client, mock_post_datapoints):
        dps = [(i * 1e11, i) for i in range(1, 11)]
        res = cognite_client.datapoints.insert(dps, id=1)
        assert res is None
        assert {
            "items": [{"id": 1, "datapoints": [{"timestamp": int(i * 1e11), "value": i} for i in range(1, 11)]}]
        } == jsgz_load(mock_post_datapoints.calls[0].request.body)

    def test_insert_dicts(self, cognite_client, mock_post_datapoints):
        dps = [{"timestamp": i * 1e11, "value": i} for i in range(1, 11)]
        res = cognite_client.datapoints.insert(dps, id=1)
        assert res is None
        assert {
            "items": [{"id": 1, "datapoints": [{"timestamp": int(i * 1e11), "value": i} for i in range(1, 11)]}]
        } == jsgz_load(mock_post_datapoints.calls[0].request.body)

    def test_by_external_id(self, cognite_client, mock_post_datapoints):
        dps = [(i * 1e11, i) for i in range(1, 11)]
        cognite_client.datapoints.insert(dps, external_id="1")
        assert {
            "items": [
                {"externalId": "1", "datapoints": [{"timestamp": int(i * 1e11), "value": i} for i in range(1, 11)]}
            ]
        } == jsgz_load(mock_post_datapoints.calls[0].request.body)

    @pytest.mark.parametrize("ts_key, value_key", [("timestamp", "values"), ("timstamp", "value")])
    def test_invalid_datapoints_keys(self, cognite_client, ts_key, value_key):
        dps = [{ts_key: i * 1e11, value_key: i} for i in range(1, 11)]
        with pytest.raises(AssertionError, match="is missing the"):
            cognite_client.datapoints.insert(dps, id=1)

    def test_insert_datapoints_over_limit(self, cognite_client, mock_post_datapoints):
        dps = [(i * 1e11, i) for i in range(1, 11)]
        with set_request_limit(cognite_client.datapoints, 5):
            res = cognite_client.datapoints.insert(dps, id=1)
        assert res is None
        request_bodies = [jsgz_load(call.request.body) for call in mock_post_datapoints.calls]
        assert {
            "items": [{"id": 1, "datapoints": [{"timestamp": int(i * 1e11), "value": i} for i in range(1, 6)]}]
        } in request_bodies
        assert {
            "items": [{"id": 1, "datapoints": [{"timestamp": int(i * 1e11), "value": i} for i in range(6, 11)]}]
        } in request_bodies

    def test_insert_datapoints_no_data(self, cognite_client):
        with pytest.raises(AssertionError, match="No datapoints provided"):
            cognite_client.datapoints.insert(id=1, datapoints=[])

    def test_insert_datapoints_in_multiple_time_series(self, cognite_client, mock_post_datapoints):
        dps = [{"timestamp": i * 1e11, "value": i} for i in range(1, 11)]
        dps_objects = [{"externalId": "1", "datapoints": dps}, {"id": 1, "datapoints": dps}]
        res = cognite_client.datapoints.insert_multiple(dps_objects)
        assert res is None
        request_body = jsgz_load(mock_post_datapoints.calls[0].request.body)
        assert {
            "items": [
                {"externalId": "1", "datapoints": [{"timestamp": int(i * 1e11), "value": i} for i in range(1, 11)]},
                {"id": 1, "datapoints": [{"timestamp": int(i * 1e11), "value": i} for i in range(1, 11)]},
            ]
        } == request_body

    def test_insert_datapoints_in_multiple_time_series_invalid_key(self, cognite_client):
        dps = [{"timestamp": i * 1e11, "value": i} for i in range(1, 11)]
        dps_objects = [{"extId": "1", "datapoints": dps}]
        with pytest.raises(ValueError, match="Invalid key 'extId'"):
            cognite_client.datapoints.insert_multiple(dps_objects)

    def test_insert_datapoints_ts_does_not_exist(self, cognite_client, mock_post_datapoints_400):
        with pytest.raises(CogniteNotFoundError):
            cognite_client.datapoints.insert(datapoints=[(1e14, 1)], external_id="does_not_exist")

    def test_insert_multiple_ts__below_ts_and_dps_limit(self, cognite_client, mock_post_datapoints):
        dps = [{"timestamp": i * 1e11, "value": i} for i in range(1, 2)]
        dps_objects = [{"id": i, "datapoints": dps} for i in range(100)]
        cognite_client.datapoints.insert_multiple(dps_objects)
        assert 1 == len(mock_post_datapoints.calls)
        request_body = jsgz_load(mock_post_datapoints.calls[0].request.body)
        for i, dps in enumerate(request_body["items"]):
            assert i == dps["id"]

    @pytest.fixture
    def set_post_dps_objects_limit_to_100(self, cognite_client):
        tmp = cognite_client.datapoints._POST_DPS_OBJECTS_LIMIT
        cognite_client.datapoints._POST_DPS_OBJECTS_LIMIT = 100
        yield
        cognite_client.datapoints._POST_DPS_OBJECTS_LIMIT = tmp

    def test_insert_multiple_ts_single_call__below_dps_limit_above_ts_limit(
        self, cognite_client, mock_post_datapoints, set_post_dps_objects_limit_to_100
    ):
        dps = [{"timestamp": i * 1e11, "value": i} for i in range(1, 2)]
        dps_objects = [{"id": i, "datapoints": dps} for i in range(101)]
        cognite_client.datapoints.insert_multiple(dps_objects)
        assert 2 == len(mock_post_datapoints.calls)

    def test_insert_multiple_ts_single_call__above_dps_limit_below_ts_limit(self, cognite_client, mock_post_datapoints):
        dps = [{"timestamp": i * 1e11, "value": i} for i in range(1, 1002)]
        dps_objects = [{"id": i, "datapoints": dps} for i in range(10)]
        cognite_client.datapoints.insert_multiple(dps_objects)
        assert 2 == len(mock_post_datapoints.calls)


@pytest.fixture
def mock_delete_datapoints(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/delete",
        status=200,
        json={},
    )
    yield rsps


class TestDeleteDatapoints:
    def test_delete_range(self, cognite_client, mock_delete_datapoints):
        res = cognite_client.datapoints.delete_range(start=datetime(2018, 1, 1), end=datetime(2018, 1, 2), id=1)
        assert res is None
        assert {"items": [{"id": 1, "inclusiveBegin": 1514764800000, "exclusiveEnd": 1514851200000}]} == jsgz_load(
            mock_delete_datapoints.calls[0].request.body
        )

    @pytest.mark.parametrize(
        "id, external_id, exception",
        [(None, None, AssertionError), (1, "1", AssertionError), ("1", None, TypeError), (None, 1, TypeError)],
    )
    def test_delete_range_invalid_id(self, cognite_client, id, external_id, exception):
        with pytest.raises(exception):
            cognite_client.datapoints.delete_range("1d-ago", "now", id, external_id)

    def test_delete_range_start_after_end(self, cognite_client):
        with pytest.raises(AssertionError, match="must be"):
            cognite_client.datapoints.delete_range(1, 0, 1)

    def test_delete_ranges(self, cognite_client, mock_delete_datapoints):
        ranges = [{"id": 1, "start": 0, "end": 1}, {"externalId": "1", "start": 0, "end": 1}]
        cognite_client.datapoints.delete_ranges(ranges)
        assert {
            "items": [
                {"id": 1, "inclusiveBegin": 0, "exclusiveEnd": 1},
                {"externalId": "1", "inclusiveBegin": 0, "exclusiveEnd": 1},
            ]
        } == jsgz_load(mock_delete_datapoints.calls[0].request.body)

    def test_delete_ranges_invalid_ids(self, cognite_client):
        ranges = [{"idz": 1, "start": 0, "end": 1}]
        with pytest.raises(AssertionError, match="Invalid key 'idz'"):
            cognite_client.datapoints.delete_ranges(ranges)
        ranges = [{"start": 0, "end": 1}]
        with pytest.raises(AssertionError, match="Exactly one of id and external id must be specified"):
            cognite_client.datapoints.delete_ranges(ranges)


class TestDatapointsObject:
    def test_len(self, cognite_client):
        assert 3 == len(Datapoints(id=1, timestamp=[1, 2, 3], value=[1, 2, 3]))

    def test_get_non_empty_data_fields(self, cognite_client):
        assert sorted([("timestamp", [1, 2, 3]), ("value", [1, 2, 3])]) == sorted(
            Datapoints(id=1, timestamp=[1, 2, 3], value=[1, 2, 3])._get_non_empty_data_fields()
        )
        assert sorted([("timestamp", [1, 2, 3]), ("max", [1, 2, 3]), ("sum", [1, 2, 3])]) == sorted(
            Datapoints(id=1, timestamp=[1, 2, 3], sum=[1, 2, 3], max=[1, 2, 3])._get_non_empty_data_fields()
        )
        assert sorted([("timestamp", [1, 2, 3]), ("max", [1, 2, 3])]) == sorted(
            Datapoints(id=1, timestamp=[1, 2, 3], sum=[], max=[1, 2, 3])._get_non_empty_data_fields()
        )
        assert sorted([("timestamp", [1, 2, 3]), ("max", [1, 2, 3]), ("sum", [])]) == sorted(
            Datapoints(id=1, timestamp=[1, 2, 3], sum=[], max=[1, 2, 3])._get_non_empty_data_fields(
                get_empty_lists=True
            )
        )
        assert [("timestamp", [])] == list(Datapoints(id=1)._get_non_empty_data_fields())

    def test_iter(self, cognite_client):
        for dp in Datapoints(id=1, timestamp=[1, 2, 3], value=[1, 2, 3]):
            assert dp.timestamp in [1, 2, 3]
            assert dp.value in [1, 2, 3]

    def test_eq(self, cognite_client):
        assert Datapoints(1) == Datapoints(1)
        assert Datapoints(1, timestamp=[1, 2, 3], value=[1, 2, 3]) == Datapoints(
            1, timestamp=[1, 2, 3], value=[1, 2, 3]
        )
        assert Datapoints(1) != Datapoints(0)
        assert Datapoints(1, timestamp=[1, 2, 3], value=[1, 2, 3]) != Datapoints(1, timestamp=[1, 2, 3], max=[1, 2, 3])
        assert Datapoints(1, timestamp=[1, 2, 3], value=[1, 2, 3]) != Datapoints(
            1, timestamp=[1, 2, 3], value=[1, 2, 4]
        )

    def test_get_item(self, cognite_client):
        dps = Datapoints(id=1, timestamp=[1, 2, 3], value=[1, 2, 3])

        assert Datapoint(timestamp=1, value=1) == dps[0]
        assert Datapoint(timestamp=2, value=2) == dps[1]
        assert Datapoint(timestamp=3, value=3) == dps[2]
        assert Datapoints(id=1, timestamp=[1, 2], value=[1, 2]) == dps[:2]

    def test_load(self, cognite_client):
        res = Datapoints._load(
            {
                "id": 1,
                "externalId": "1",
                "isString": False,
                "isStep": False,
                "unit": "kPa",
                "datapoints": [{"timestamp": 1, "value": 1}, {"timestamp": 2, "value": 2}],
            }
        )
        assert 1 == res.id
        assert "1" == res.external_id
        assert [1, 2] == res.timestamp
        assert [1, 2] == res.value
        assert "kPa" == res.unit
        assert res.is_step is False
        assert res.is_string is False

    def test_load_string(self, cognite_client):
        res = Datapoints._load(
            {
                "id": 1,
                "externalId": "1",
                "isString": True,
                "datapoints": [{"timestamp": 1, "value": 1}, {"timestamp": 2, "value": 2}],
            }
        )
        assert 1 == res.id
        assert "1" == res.external_id
        assert [1, 2] == res.timestamp
        assert [1, 2] == res.value
        assert res.is_string is True
        assert res.is_step is None
        assert res.unit is None

    def test_slice(self, cognite_client):
        res = Datapoints(id=1, timestamp=[1, 2, 3])._slice(slice(None, 1))
        assert [1] == res.timestamp

    def test_extend(self, cognite_client):
        d0 = Datapoints()
        d1 = Datapoints(id=1, external_id="1", timestamp=[1, 2, 3], value=[1, 2, 3])
        d2 = Datapoints(id=1, external_id="1", timestamp=[4, 5, 6], value=[4, 5, 6])
        d3 = Datapoints(id=1, external_id="1", timestamp=[7, 8, 9, 10], value=[7, 8, 9, 10])

        d0._extend(d1)
        assert [1, 2, 3] == d0.timestamp
        assert [1, 2, 3] == d0.value
        assert 1 == d0.id
        assert "1" == d0.external_id
        assert d0.sum is None

        d0._extend(d2)
        assert [1, 2, 3, 4, 5, 6] == d0.value
        assert [1, 2, 3, 4, 5, 6] == d0.timestamp
        assert 1 == d0.id
        assert "1" == d0.external_id
        assert d0.sum is None

        d0._extend(d3)
        assert [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] == d0.timestamp
        assert [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] == d0.value
        assert 1 == d0.id
        assert "1" == d0.external_id
        assert d0.sum is None


@pytest.mark.dsl
class TestPlotDatapoints:
    @mock.patch("matplotlib.pyplot.show")
    @mock.patch("pandas.core.frame.DataFrame.plot")
    def test_plot_datapoints(self, pandas_plot_mock, plt_show_mock):
        d = Datapoints(id=1, timestamp=[1, 2, 3, 4, 5], value=[1, 2, 3, 4, 5])
        d.plot()
        assert 1 == pandas_plot_mock.call_count
        assert 1 == plt_show_mock.call_count

    @mock.patch("matplotlib.pyplot.show")
    @mock.patch("pandas.core.frame.DataFrame.plot")
    def test_plot_datapoints_list(self, pandas_plot_mock, plt_show_mock):
        d1 = Datapoints(id=1, timestamp=[1, 2, 3, 4, 5], value=[1, 2, 3, 4, 5])
        d2 = Datapoints(id=2, timestamp=[1, 2, 3, 4, 5], value=[6, 7, 8, 9, 10])
        d = DatapointsList([d1, d2])
        d.plot()
        assert 1 == pandas_plot_mock.call_count
        assert 1 == plt_show_mock.call_count


@pytest.mark.dsl
class TestPandasIntegration:
    def test_datapoint(self, cognite_client):
        import pandas as pd

        d = Datapoint(timestamp=0, value=2, max=3)
        expected_df = pd.DataFrame({"value": [2], "max": [3]}, index=[utils._time.ms_to_datetime(0)])
        pd.testing.assert_frame_equal(expected_df, d.to_pandas(), check_like=True)

    def test_datapoints(self, cognite_client):
        import pandas as pd

        d = Datapoints(id=1, timestamp=[1, 2, 3], average=[2, 3, 4], step_interpolation=[3, 4, 5])
        expected_df = pd.DataFrame(
            {"1|average": [2, 3, 4], "1|stepInterpolation": [3, 4, 5]},
            index=[utils._time.ms_to_datetime(ms) for ms in [1, 2, 3]],
        )
        pd.testing.assert_frame_equal(expected_df, d.to_pandas())

    def test_datapoints_no_names(self, cognite_client):
        import pandas as pd

        d = Datapoints(id=1, timestamp=[1, 2, 3], average=[2, 3, 4])
        expected_df = pd.DataFrame({"1": [2, 3, 4]}, index=[utils._time.ms_to_datetime(ms) for ms in [1, 2, 3]])
        pd.testing.assert_frame_equal(expected_df, d.to_pandas(include_aggregate_name=False))
        expected_df = pd.DataFrame({"1|average": [2, 3, 4]}, index=[utils._time.ms_to_datetime(ms) for ms in [1, 2, 3]])
        pd.testing.assert_frame_equal(expected_df, d.to_pandas(include_aggregate_name=True))

    def test_id_and_external_id_set_gives_external_id_columns(self, cognite_client):
        import pandas as pd

        d = Datapoints(id=0, external_id="abc", timestamp=[1, 2, 3], average=[2, 3, 4], step_interpolation=[3, 4, 5])
        expected_df = pd.DataFrame(
            {"abc|average": [2, 3, 4], "abc|stepInterpolation": [3, 4, 5]},
            index=[utils._time.ms_to_datetime(ms) for ms in [1, 2, 3]],
        )
        pd.testing.assert_frame_equal(expected_df, d.to_pandas())

    def test_datapoints_empty(self, cognite_client):
        d = Datapoints(external_id="1", timestamp=[], value=[])
        assert d.to_pandas().empty

    def test_datapoints_list(self, cognite_client):
        import pandas as pd

        d1 = Datapoints(id=1, timestamp=[1, 2, 3], average=[2, 3, 4], step_interpolation=[3, 4, 5])
        d2 = Datapoints(id=2, timestamp=[1, 2, 3], max=[2, 3, 4], step_interpolation=[3, 4, 5])
        d3 = Datapoints(id=3, timestamp=[1, 3], value=[1, 3])
        dps_list = DatapointsList([d1, d2, d3])
        expected_df = pd.DataFrame(
            {
                "1|average": [2, 3, 4],
                "1|stepInterpolation": [3, 4, 5],
                "2|max": [2, 3, 4],
                "2|stepInterpolation": [3, 4, 5],
                "3": [1, None, 3],
            },
            index=[utils._time.ms_to_datetime(ms) for ms in [1, 2, 3]],
        )
        pd.testing.assert_frame_equal(expected_df, dps_list.to_pandas())

    def test_datapoints_list_names(self, cognite_client):
        import pandas as pd

        d1 = Datapoints(id=2, timestamp=[1, 2, 3], max=[2, 3, 4])
        d2 = Datapoints(id=3, timestamp=[1, 3], average=[1, 3])
        dps_list = DatapointsList([d1, d2])
        expected_df = pd.DataFrame(
            {"2|max": [2, 3, 4], "3|average": [1, None, 3]}, index=[utils._time.ms_to_datetime(ms) for ms in [1, 2, 3]]
        )
        pd.testing.assert_frame_equal(expected_df, dps_list.to_pandas())
        expected_df.columns = [c[:1] for c in expected_df.columns]
        pd.testing.assert_frame_equal(expected_df, dps_list.to_pandas(include_aggregate_name=False))

    def test_datapoints_list_names_dup(self, cognite_client):
        import pandas as pd

        d1 = Datapoints(id=2, timestamp=[1, 2, 3], max=[2, 3, 4])
        d2 = Datapoints(id=2, timestamp=[1, 3], average=[1, 3])
        dps_list = DatapointsList([d1, d2])
        expected_df = pd.DataFrame(
            {"2|max": [2, 3, 4], "2|average": [1, None, 3]},
            index=[utils._time.ms_to_datetime(ms) for ms in [1, 2, 3]],
            columns=["2|max", "2|average"],
        )
        pd.testing.assert_frame_equal(expected_df, dps_list.to_pandas())
        with pytest.raises(CogniteDuplicateColumnsError):
            dps_list.to_pandas(include_aggregate_name=False)

    def test_datapoints_list_non_aligned(self, cognite_client):
        import pandas as pd

        d1 = Datapoints(id=1, timestamp=[1, 2, 3], value=[1, 2, 3])
        d2 = Datapoints(id=2, timestamp=[3, 4, 5], value=[3, 4, 5])

        dps_list = DatapointsList([d1, d2])

        expected_df = pd.DataFrame(
            {"1": [1, 2, 3, None, None], "2": [None, None, 3, 4, 5]},
            index=[utils._time.ms_to_datetime(ms) for ms in [1, 2, 3, 4, 5]],
        )
        pd.testing.assert_frame_equal(expected_df, dps_list.to_pandas())

    def test_datapoints_list_empty(self, cognite_client):
        dps_list = DatapointsList([])
        assert dps_list.to_pandas().empty

    def test_retrieve_dataframe(self, cognite_client, mock_get_datapoints):
        df = cognite_client.datapoints.retrieve_dataframe(
            id=[1, {"id": 2, "aggregates": ["max"]}],
            external_id=["123"],
            start=1000000,
            end=1100000,
            aggregates=["average"],
            granularity="10s",
        )

        assert {"1|average", "2|max", "123|average"} == set(df.columns)
        assert df.shape[0] > 0

    def test_retrieve_datapoints_some_aggregates_omitted(
        self, cognite_client, mock_get_datapoints_one_ts_has_missing_aggregates
    ):
        import pandas as pd

        df = cognite_client.datapoints.retrieve_dataframe(
            id={"id": 1, "aggregates": ["average"]},
            external_id={"externalId": "def", "aggregates": ["interpolation"]},
            start=0,
            end=1,
            aggregates=[],
            granularity="1s",
        )

        expected_df = pd.DataFrame(
            {"1|average": [0, 1, 2, 3, 4], "def|interpolation": [None, 1, None, 3, None]},
            index=[utils._time.ms_to_datetime(i) for i in range(5)],
        )
        pd.testing.assert_frame_equal(df, expected_df)

    def test_retrieve_datapoints_last_beyond_end(self, cognite_client, mock_get_datapoints_include_outside):
        dpt = cognite_client.datapoints.retrieve(
            id=1, include_outside_points=True, start=1000000000, end=1000000000 + 100000
        )
        assert 100001 == len(dpt)

    def test_retrieve_dataframe_several_missing(self, cognite_client, mock_get_datapoints_several_missing):
        import pandas as pd

        df = cognite_client.datapoints.retrieve_dataframe(
            id=[
                {"id": 1, "aggregates": ["average"]},
                {"id": 2, "aggregates": ["interpolation"]},
                {"id": 3, "aggregates": ["count"]},
            ],
            aggregates=[],
            start=0,
            end=1,
            granularity="1s",
        )

        expected_df = pd.DataFrame(
            {"1|average": [11, 22, 44, None], "2|interpolation": [None, 1, 3, None], "3|count": [None, 2, 4, 5]},
            index=[utils._time.ms_to_datetime(i * 1000) for i in [0, 1, 3, 4]],
        )
        pd.testing.assert_frame_equal(df, expected_df)

    def test_retrieve_dataframe_complete_single_isstep(self, cognite_client, mock_get_datapoints_single_isstep):
        import pandas as pd

        df = cognite_client.datapoints.retrieve_dataframe(
            id=[{"id": 3, "aggregates": ["interpolation"]}],
            aggregates=[],
            start=0,
            end=1,
            granularity="1s",
            complete="fill",
            include_aggregate_name=False,
        )

        expected_df = pd.DataFrame(
            {"3": [1.0, 1.0, 3.0]}, index=[utils._time.ms_to_datetime(i * 1000) for i in [1, 2, 3]]
        )
        pd.testing.assert_frame_equal(df, expected_df)

    def test_retrieve_dataframe_several_missing_complete(self, cognite_client, mock_get_datapoints_several_missing):
        import pandas as pd

        df = cognite_client.datapoints.retrieve_dataframe(
            id=[
                {"id": 1, "aggregates": ["average"]},
                {"id": 2, "aggregates": ["interpolation"]},
                {"id": 3, "aggregates": ["count"]},
            ],
            aggregates=[],
            start=0,
            end=1,
            granularity="1s",
            complete="fill",
        )

        expected_df = pd.DataFrame(
            {
                "1|average": [11.0, 22.0, None, 44.0, None],
                "2|interpolation": [None, 1.0, 2.0, 3.0, None],
                "3|count": [0.0, 2.0, 0.0, 4.0, 5.0],
            },
            index=[utils._time.ms_to_datetime(i * 1000) for i in range(5)],
        )
        pd.testing.assert_frame_equal(df, expected_df)

    def test_retrieve_dataframe_dict_empty(self, cognite_client, mock_get_datapoints_empty):
        dfd = cognite_client.datapoints.retrieve_dataframe_dict(
            id=1,
            aggregates=["count", "interpolation", "stepInterpolation", "totalVariation"],
            start=0,
            end=1,
            granularity="1s",
        )
        assert isinstance(dfd, dict)
        assert 4 == len(dfd)

    def test_retrieve_dataframe_dict_empty_single_aggregate(self, cognite_client, mock_get_datapoints_empty):
        dfd = cognite_client.datapoints.retrieve_dataframe_dict(
            id=1, aggregates=["count"], start=0, end=1, granularity="1s"
        )
        assert isinstance(dfd, dict)
        assert ["count"] == list(dfd.keys())
        assert dfd["count"].empty

    def test_retrieve_dataframe_complete_all(self, cognite_client, mock_get_datapoints):
        import pandas as pd

        df = cognite_client.datapoints.retrieve_dataframe(
            id=[1, 2],
            aggregates=["count", "sum", "average", "totalVariation"],
            start=0,
            end=1,
            granularity="1s",
            complete="fill",
        )
        assert isinstance(df, pd.DataFrame)
        assert 8 == df.shape[1]

    def test_retrieve_dataframe_dict(self, cognite_client, mock_get_datapoints_several_missing):
        import pandas as pd

        dfd = cognite_client.datapoints.retrieve_dataframe_dict(
            id=[
                {"id": 1, "aggregates": ["average"]},
                {"id": 2, "aggregates": ["interpolation"]},
                {"id": 3, "aggregates": ["count"]},
            ],
            aggregates=[],
            start=0,
            end=1,
            granularity="1s",
        )
        assert isinstance(dfd, dict)
        assert 3 == len(dfd)

        expected_dict = {
            "average": pd.DataFrame(
                {"1": [11.0, 22.0, 44.0, None]}, index=[utils._time.ms_to_datetime(i * 1000) for i in [0, 1, 3, 4]]
            ),
            "count": pd.DataFrame(
                {"3": [None, 2, 4, 5]}, index=[utils._time.ms_to_datetime(i * 1000) for i in [0, 1, 3, 4]]
            ),
            "interpolation": pd.DataFrame(
                {"2": [None, 1, 3, None]}, index=[utils._time.ms_to_datetime(i * 1000) for i in [0, 1, 3, 4]]
            ),
        }
        for k in expected_dict:
            pd.testing.assert_frame_equal(expected_dict[k], dfd[k])

    def test_retrieve_dataframe_dict_complete(self, cognite_client, mock_get_datapoints_several_missing):
        import pandas as pd

        dfd = cognite_client.datapoints.retrieve_dataframe_dict(
            id=[{"id": 2, "aggregates": ["interpolation"]}, {"id": 3, "aggregates": ["count"]}],
            aggregates=[],
            start=0,
            end=1,
            granularity="1s",
            complete="fill,dropna",
        )
        assert isinstance(dfd, dict)
        assert 2 == len(dfd)

        expected_dict = {
            "count": pd.DataFrame(
                {"3": [2.0, 0.0, 4.0]}, index=[utils._time.ms_to_datetime(i * 1000) for i in [1, 2, 3]]
            ),
            "interpolation": pd.DataFrame(
                {"2": [1.0, 2.0, 3.0]}, index=[utils._time.ms_to_datetime(i * 1000) for i in [1, 2, 3]]
            ),
        }

        for k in expected_dict:
            pd.testing.assert_frame_equal(expected_dict[k], dfd[k])

        with pytest.raises(ValueError, match="is not supported for dataframe completion"):
            dfd = cognite_client.datapoints.retrieve_dataframe_dict(
                id=[{"id": 1, "aggregates": ["average"]}],
                aggregates=[],
                start=0,
                end=1,
                granularity="1s",
                complete="fill,dropna",
            )

    def test_retrieve_dataframe_id_and_external_id_requested(self, cognite_client, rsps):
        rsps.add(
            rsps.POST,
            cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
            status=200,
            json={
                "items": [
                    {
                        "id": 1,
                        "externalId": "abc",
                        "isString": False,
                        "isStep": False,
                        "datapoints": [{"timestamp": 0, "average": 1}],
                    }
                ]
            },
        )
        rsps.add(
            rsps.POST,
            cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
            status=200,
            json={
                "items": [
                    {
                        "id": 2,
                        "externalId": "def",
                        "isString": False,
                        "isStep": False,
                        "datapoints": [{"timestamp": 0, "average": 1}],
                    }
                ]
            },
        )
        res = cognite_client.datapoints.retrieve_dataframe(
            start=0, end="now", id=1, external_id=["def"], aggregates=["average"], granularity="1m"
        )
        assert {"1|average", "def|average"} == set(res.columns)

    def test_insert_dataframe(self, cognite_client, mock_post_datapoints):
        import pandas as pd

        timestamps = [1500000000000, 1510000000000, 1520000000000, 1530000000000]
        df = pd.DataFrame(
            {"123": [1, 2, 3, 4], "456": [5.0, 6.0, 7.0, 8.0]},
            index=[utils._time.ms_to_datetime(ms) for ms in timestamps],
        )
        res = cognite_client.datapoints.insert_dataframe(df)
        assert res is None
        request_body = jsgz_load(mock_post_datapoints.calls[0].request.body)
        assert {
            "items": [
                {
                    "id": 123,
                    "datapoints": [{"timestamp": ts, "value": val} for ts, val in zip(timestamps, range(1, 5))],
                },
                {
                    "id": 456,
                    "datapoints": [{"timestamp": ts, "value": float(val)} for ts, val in zip(timestamps, range(5, 9))],
                },
            ]
        } == request_body

    def test_insert_dataframe_external_ids(self, cognite_client, mock_post_datapoints):
        import pandas as pd

        timestamps = [1500000000000, 1510000000000, 1520000000000, 1530000000000]
        df = pd.DataFrame(
            {"123": [1, 2, 3, 4], "456": [5.0, 6.0, 7.0, 8.0]},
            index=[utils._time.ms_to_datetime(ms) for ms in timestamps],
        )
        res = cognite_client.datapoints.insert_dataframe(df, external_id_headers=True)
        assert res is None
        request_body = jsgz_load(mock_post_datapoints.calls[0].request.body)
        assert {
            "items": [
                {
                    "externalId": "123",
                    "datapoints": [{"timestamp": ts, "value": val} for ts, val in zip(timestamps, range(1, 5))],
                },
                {
                    "externalId": "456",
                    "datapoints": [{"timestamp": ts, "value": float(val)} for ts, val in zip(timestamps, range(5, 9))],
                },
            ]
        } == request_body

    def test_insert_dataframe_with_nans(self, cognite_client):
        import pandas as pd

        timestamps = [1500000000000, 1510000000000, 1520000000000, 1530000000000]
        df = pd.DataFrame(
            {"123": [1, 2, None, 4], "456": [5.0, 6.0, 7.0, 8.0]},
            index=[utils._time.ms_to_datetime(ms) for ms in timestamps],
        )
        with pytest.raises(AssertionError, match="contains NaNs"):
            cognite_client.datapoints.insert_dataframe(df)

    def test_insert_dataframe_with_dropna(self, cognite_client, mock_post_datapoints):
        import pandas as pd

        timestamps = [1500000000000, 1510000000000, 1520000000000, 1530000000000]
        df = pd.DataFrame(
            {"123": [1, 2, None, 4], "456": [5.0, 6.0, 7.0, 8.0]},
            index=[utils._time.ms_to_datetime(ms) for ms in timestamps],
        )
        res = cognite_client.datapoints.insert_dataframe(df, dropna=True)
        assert res is None
        request_body = jsgz_load(mock_post_datapoints.calls[0].request.body)
        assert {
            "items": [
                {
                    "id": 123,
                    "datapoints": [
                        {"timestamp": ts, "value": val} for ts, val in zip(timestamps, range(1, 5)) if val != 3
                    ],
                },
                {
                    "id": 456,
                    "datapoints": [{"timestamp": ts, "value": float(val)} for ts, val in zip(timestamps, range(5, 9))],
                },
            ]
        } == request_body

    def test_insert_dataframe_single_dp(self, cognite_client, mock_post_datapoints):
        import pandas as pd

        timestamps = [1500000000000]
        df = pd.DataFrame({"a": [1.0], "b": [2.0]}, index=[utils._time.ms_to_datetime(ms) for ms in timestamps])
        res = cognite_client.datapoints.insert_dataframe(df, external_id_headers=True)
        assert res is None

    def test_insert_dataframe_with_infs(self, cognite_client):
        import numpy as np
        import pandas as pd

        timestamps = [1500000000000, 1510000000000, 1520000000000, 1530000000000]
        df = pd.DataFrame(
            {"123": [1, 2, np.inf, 4], "456": [5.0, 6.0, 7.0, 8.0], "xyz": ["a", "b", "c", "d"]},
            index=[utils._time.ms_to_datetime(ms) for ms in timestamps],
        )
        with pytest.raises(AssertionError, match="contains Infinity"):
            cognite_client.datapoints.insert_dataframe(df)

    def test_insert_dataframe_with_strings(self, cognite_client, mock_post_datapoints):
        import pandas as pd

        timestamps = [1500000000000, 1510000000000, 1520000000000, 1530000000000]
        df = pd.DataFrame(
            {"123": ["a", "b", "c", "d"], "456": [5.0, 6.0, 7.0, 8.0]},
            index=[utils._time.ms_to_datetime(ms) for ms in timestamps],
        )
        cognite_client.datapoints.insert_dataframe(df)

    def test_retrieve_datapoints_multiple_time_series_correct_ordering(self, cognite_client, mock_get_datapoints):
        ids = [1, 2, 3]
        external_ids = ["4", "5", "6"]
        dps_res_list = cognite_client.datapoints.retrieve(id=ids, external_id=external_ids, start=0, end=100000)
        assert list(dps_res_list.to_pandas().columns) == ["1", "2", "3", "4", "5", "6"], "Incorrect column ordering"

    def test_retrieve_datapoints_one_ts_empty_correct_number_of_columns(
        self, cognite_client, mock_get_datapoints_one_ts_empty
    ):
        res = cognite_client.datapoints.retrieve(id=[1, 2], start=0, end=10000)
        assert 2 == len(res.to_pandas().columns)


@pytest.fixture
def mock_get_dps_count(rsps, cognite_client):
    def request_callback(request):
        payload = jsgz_load(request.body)
        granularity = payload["granularity"]
        start = payload["start"]
        end = payload["end"]

        assert payload["aggregates"] == ["count"]
        assert utils._time.granularity_to_ms(payload["granularity"]) >= utils._time.granularity_to_ms("1d")

        dps = [{"timestamp": i, "count": 1000} for i in range(start, end, utils._time.granularity_to_ms(granularity))]
        response = {"items": [{"id": 0, "externalId": "bla", "isStep": False, "isString": False, "datapoints": dps}]}
        return 200, {}, json.dumps(response)

    rsps.add_callback(
        rsps.POST,
        cognite_client.datapoints._get_base_url_with_base_path() + "/timeseries/data/list",
        callback=request_callback,
        content_type="application/json",
    )
    yield rsps


class TestDataPoster:
    def test_datapoints_bin_add_dps_object(self, cognite_client):
        bin = DatapointsBin(10, 10)
        dps_object = {"id": 100, "datapoints": [{"timestamp": 1, "value": 1}]}
        bin.add(dps_object)
        assert 1 == bin.current_num_datapoints
        assert [dps_object] == bin.dps_object_list

    def test_datapoints_bin_will_fit__below_dps_and_ts_limit(self, cognite_client):
        bin = DatapointsBin(10, 10)
        assert bin.will_fit(10)
        assert not bin.will_fit(11)

    def test_datapoints_bin_will_fit__below_dps_limit_above_ts_limit(self, cognite_client):
        bin = DatapointsBin(1, 100)
        dps_object = {"id": 100, "datapoints": [{"timestamp": 1, "value": 1}]}
        bin.add(dps_object)
        assert not bin.will_fit(10)

    def test_datapoints_bin_will_fit__above_dps_limit_above_ts_limit(self, cognite_client):
        bin = DatapointsBin(1, 1)
        dps_object = {"id": 100, "datapoints": [{"timestamp": 1, "value": 1}]}
        bin.add(dps_object)
        assert not bin.will_fit(1)


gms = utils._time.granularity_to_ms


class TestDataFetcher:
    @pytest.mark.parametrize(
        "q, exc, message",
        [
            (
                DatapointsQuery(start=1, end=2, id=1, aggregates=["average"]),
                ValueError,
                "granularity must also be provided",
            ),
            (DatapointsQuery(start=1, end=2, id=1, granularity="1d"), ValueError, "aggregates must also be provided"),
            (
                DatapointsQuery(start=1, end=2, id=[1, 1], granularity="1d", aggregates=["average"]),
                ValueError,
                "identifier '1' is duplicated in query",
            ),
            (
                DatapointsQuery(
                    start=1, end=2, id=[1, {"id": 1, "aggregates": ["max"]}], granularity="1d", aggregates=["average"]
                ),
                ValueError,
                "identifier '1' is duplicated in query",
            ),
        ],
    )
    def test_validate_query(self, cognite_client, q, exc, message):
        with pytest.raises(exc, match=message):
            DatapointsFetcher(cognite_client.datapoints).fetch(q)

    @pytest.mark.parametrize(
        "fn",
        [
            lambda cognite_client: (
                [_DPTask(cognite_client.datapoints, 1, 2, {}, None, None, None, None, False)],
                [_DPTask(cognite_client.datapoints, 1, 2, {}, None, None, None, None, False)],
            ),
            lambda cognite_client: (
                [
                    _DPTask(
                        cognite_client.datapoints,
                        datetime(2018, 1, 1),
                        datetime(2019, 1, 1),
                        {},
                        None,
                        None,
                        None,
                        None,
                        False,
                    )
                ],
                [_DPTask(cognite_client.datapoints, 1514764800000, 1546300800000, {}, None, None, None, None, False)],
            ),
            lambda cognite_client: (
                [_DPTask(cognite_client.datapoints, gms("1h"), gms(("25h")), {}, ["average"], "1d", None, None, False)],
                [_DPTask(cognite_client.datapoints, gms("1d"), gms("2d"), {}, ["average"], "1d", None, None, False)],
            ),
        ],
    )
    def test_preprocess_tasks(self, cognite_client, fn):
        q, expected_q = fn(cognite_client)
        DatapointsFetcher(cognite_client.datapoints)._preprocess_tasks(q)
        for actual, expected in zip(q, expected_q):
            assert expected.start == actual.start
            assert expected.end == actual.end

    @pytest.mark.parametrize(
        "ts, granularity, expected_output",
        [
            (gms("1h"), "1d", gms("1d")),
            (gms("23h"), "10d", gms("1d")),
            (gms("24h"), "5d", gms("1d")),
            (gms("25h"), "3d", gms("2d")),
            (gms("1m"), "1h", gms("1h")),
            (gms("90s"), "10m", gms("2m")),
            (gms("90s"), "1s", gms("90s")),
        ],
    )
    def test_align_with_granularity_unit(self, cognite_client, ts, granularity, expected_output):
        assert expected_output == DatapointsFetcher._align_with_granularity_unit(ts, granularity)

    @pytest.mark.parametrize(
        "start, end, granularity, request_limit, user_limit, expected_output",
        [
            (0, gms("20d"), "10d", 2, None, [_DPWindow(0, 1728000000)]),
            (0, gms("20d"), "10d", 1, None, [_DPWindow(0, 864000000), _DPWindow(864000000, 1728000000)]),
            (
                0,
                gms("6d"),
                "1s",
                2000,
                None,
                [_DPWindow(0, gms("2d")), _DPWindow(gms("2d"), gms("4d")), _DPWindow(gms("4d"), gms("6d"))],
            ),
            (
                0,
                gms("3d"),
                None,
                1000,
                None,
                [_DPWindow(0, gms("1d")), _DPWindow(gms("1d"), gms("2d")), _DPWindow(gms("2d"), gms("3d"))],
            ),
            (0, gms("1h"), None, 2000, None, [_DPWindow(0, 3600000)]),
            (0, gms("1s"), None, 1, None, [_DPWindow(0, 1000)]),
            (0, gms("1s"), None, 1, None, [_DPWindow(0, 1000)]),
            (0, gms("3d"), None, 1000, 500, [_DPWindow(0, gms("1d"))]),
        ],
    )
    def test_get_datapoints_windows(
        self, cognite_client, start, end, granularity, request_limit, user_limit, expected_output, mock_get_dps_count
    ):
        user_limit = user_limit or float("inf")
        task = _DPTask(
            client=cognite_client.datapoints,
            start=start,
            end=end,
            ts_item={},
            granularity=granularity,
            aggregates=[],
            limit=None,
            include_outside_points=False,
            ignore_unknown_ids=False,
        )
        task.request_limit = request_limit
        res = DatapointsFetcher(cognite_client.datapoints)._get_windows(
            id=0, task=task, remaining_user_limit=user_limit
        )
        for w in expected_output:
            w.limit = user_limit
        assert expected_output == res

    @pytest.mark.parametrize(
        "start, end, granularity, expected_output",
        [
            (0, 10001, "1s", 10000),
            (0, 10000, "1m", 0),
            (0, 110000, "1m", gms("1m")),
            (0, gms("10d") - 1, "2d", gms("8d")),
            (0, gms("10d") - 1, "1m", gms("10d") - gms("1m")),
            (0, 10000, "1s", 10000),
        ],
    )
    def test_align_window_end(self, cognite_client, start, end, granularity, expected_output):
        assert expected_output == DatapointsFetcher._align_window_end(start, end, granularity)

    @pytest.mark.parametrize(
        "ids, external_ids, expected_output",
        [
            (1, None, ([{"id": 1}], True)),
            (None, "1", ([{"externalId": "1"}], True)),
            (1, "1", ([{"id": 1}, {"externalId": "1"}], False)),
            ([1], ["1"], ([{"id": 1}, {"externalId": "1"}], False)),
            ([1], None, ([{"id": 1}], False)),
            ({"id": 1, "aggregates": ["average"]}, None, ([{"id": 1, "aggregates": ["average"]}], True)),
            ({"id": 1}, {"externalId": "1"}, ([{"id": 1}, {"externalId": "1"}], False)),
            (
                [{"id": 1, "aggregates": ["average"]}],
                [{"externalId": "1", "aggregates": ["average", "sum"]}],
                ([{"id": 1, "aggregates": ["average"]}, {"externalId": "1", "aggregates": ["average", "sum"]}], False),
            ),
        ],
    )
    def test_process_time_series_input_ok(self, cognite_client, ids, external_ids, expected_output):
        assert expected_output == DatapointsFetcher._process_ts_identifiers(ids, external_ids)

    @pytest.mark.parametrize(
        "ids, external_ids, exception, match",
        [
            (1.0, None, TypeError, "Invalid type '<class 'float'>'"),
            ([1.0], None, TypeError, "Invalid type '<class 'float'>'"),
            (None, 1, TypeError, "Invalid type '<class 'int'>'"),
            (None, [1], TypeError, "Invalid type '<class 'int'>'"),
            ({"wrong": 1, "aggregates": ["average"]}, None, ValueError, "Unknown key 'wrong'"),
            (None, [{"externalId": 1, "wrong": ["average"]}], ValueError, "Unknown key 'wrong'"),
            (None, {"id": 1, "aggregates": ["average"]}, ValueError, "Unknown key 'id'"),
            ({"externalId": 1}, None, ValueError, "Unknown key 'externalId'"),
        ],
    )
    def test_process_time_series_input_fail(self, cognite_client, ids, external_ids, exception, match):
        with pytest.raises(exception, match=match):
            DatapointsFetcher._process_ts_identifiers(ids, external_ids)



from datetime import datetime
from unittest import mock

import pytest

from cognite.client.data_classes import TimeSeries, TimeSeriesFilter, TimeSeriesUpdate
from tests.utils import set_request_limit


@pytest.fixture(scope="class")
def new_ts(cognite_client):
    ts = cognite_client.time_series.create(TimeSeries(name="any", metadata={"a": "b"}))
    yield ts
    cognite_client.time_series.delete(id=ts.id)
    assert cognite_client.time_series.retrieve(ts.id) is None


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(cognite_client.time_series, "_post", wraps=cognite_client.time_series._post) as _:
        yield


class TestTimeSeriesAPI:
    def test_retrieve(self, cognite_client):
        listed_asset = cognite_client.time_series.list(limit=1)[0]
        retrieved_asset = cognite_client.time_series.retrieve(listed_asset.id)
        retrieved_asset.external_id = listed_asset.external_id
        assert retrieved_asset == listed_asset

    def test_retrieve_multiple(self, cognite_client):
        res = cognite_client.time_series.list(limit=2)
        retrieved_assets = cognite_client.time_series.retrieve_multiple([t.id for t in res])
        for listed_asset, retrieved_asset in zip(res, retrieved_assets):
            retrieved_asset.external_id = listed_asset.external_id
        assert res == retrieved_assets

    def test_retrieve_multiple_ignore_unknown(self, cognite_client):
        res = cognite_client.time_series.list(limit=2)
        retrieved_assets = cognite_client.time_series.retrieve_multiple(
            [t.id for t in res] + [1, 2, 3], ignore_unknown_ids=True
        )
        for listed_asset, retrieved_asset in zip(res, retrieved_assets):
            retrieved_asset.external_id = listed_asset.external_id
        assert res == retrieved_assets

    def test_list(self, cognite_client, post_spy):
        with set_request_limit(cognite_client.time_series, 10):
            res = cognite_client.time_series.list(limit=20)

        assert 20 == len(res)
        assert 2 == cognite_client.time_series._post.call_count

    def test_list_with_filters(self, cognite_client, post_spy):
        res = cognite_client.time_series.list(
            is_step=True,
            is_string=False,
            metadata={"a": "b"},
            last_updated_time={"min": 45},
            created_time={"max": 123},
            asset_ids=[1, 2],
            root_asset_ids=[1231],
            include_metadata=False,
        )
        assert 0 == len(res)
        assert 1 == cognite_client.time_series._post.call_count

    def test_partitioned_list(self, cognite_client, post_spy):
        mintime = datetime(2019, 1, 1).timestamp() * 1000
        maxtime = datetime(2019, 5, 15).timestamp() * 1000
        res_flat = cognite_client.time_series.list(limit=None, created_time={"min": mintime, "max": maxtime})
        res_part = cognite_client.time_series.list(
            partitions=8, limit=None, created_time={"min": mintime, "max": maxtime}
        )
        assert len(res_flat) > 0
        assert len(res_flat) == len(res_part)
        assert {a.id for a in res_flat} == {a.id for a in res_part}

    def test_aggregate(self, cognite_client, new_ts):
        res = cognite_client.time_series.aggregate(filter=TimeSeriesFilter(name="any"))
        assert res[0].count > 0

    def test_search(self, cognite_client):
        res = cognite_client.time_series.search(
            name="test__timestamp_multiplied", filter=TimeSeriesFilter(created_time={"min": 0})
        )
        assert len(res) > 0

    def test_update(self, cognite_client, new_ts):
        assert new_ts.metadata == {"a": "b"}
        update_ts = TimeSeriesUpdate(new_ts.id).name.set("newname").metadata.set({})
        res = cognite_client.time_series.update(update_ts)
        assert "newname" == res.name
        assert res.metadata == {}

    def test_delete_with_nonexisting(self, cognite_client):
        a = cognite_client.time_series.create(TimeSeries(name="any"))
        cognite_client.assets.delete(id=a.id, external_id="this ts does not exist", ignore_unknown_ids=True)
        assert cognite_client.assets.retrieve(id=a.id) is None



from unittest import mock
from unittest.mock import call

import pytest

from cognite.client.data_classes import Asset, AssetList, Event, EventList, FileMetadata, FileMetadataList


class TestAsset:
    def test_get_events(self, cognite_client):
        cognite_client.events.list = mock.MagicMock()
        a = Asset(id=1, cognite_client=cognite_client)
        a.events()
        assert cognite_client.events.list.call_args == call(asset_ids=[1])
        assert cognite_client.events.list.call_count == 1

    def test_get_time_series(self, cognite_client):
        cognite_client.time_series.list = mock.MagicMock()
        a = Asset(id=1, cognite_client=cognite_client)
        a.time_series()
        assert cognite_client.time_series.list.call_args == call(asset_ids=[1])
        assert cognite_client.time_series.list.call_count == 1

    def test_get_sequences(self, cognite_client):
        cognite_client.sequences.list = mock.MagicMock()
        a = Asset(id=1, cognite_client=cognite_client)
        a.sequences()
        assert cognite_client.sequences.list.call_args == call(asset_ids=[1])
        assert cognite_client.sequences.list.call_count == 1

    def test_get_files(self, cognite_client):
        cognite_client.files.list = mock.MagicMock()
        a = Asset(id=1, cognite_client=cognite_client)
        a.files()
        assert cognite_client.files.list.call_args == call(asset_ids=[1])
        assert cognite_client.files.list.call_count == 1

    def test_get_parent(self, cognite_client):
        cognite_client.assets.retrieve = mock.MagicMock()
        a1 = Asset(parent_id=1, cognite_client=cognite_client)
        a1.parent()
        assert cognite_client.assets.retrieve.call_args == call(id=1)
        assert cognite_client.assets.retrieve.call_count == 1

    def test_get_children(self, cognite_client):
        cognite_client.assets.list = mock.MagicMock()
        a1 = Asset(id=1, cognite_client=cognite_client)
        a1.children()
        assert cognite_client.assets.list.call_args == call(parent_ids=[1], limit=None)
        assert cognite_client.assets.list.call_count == 1

    def test_get_subtree(self, cognite_client):
        cognite_client.assets.retrieve_subtree = mock.MagicMock()
        a1 = Asset(id=1, cognite_client=cognite_client)
        a1.subtree(depth=1)
        assert cognite_client.assets.retrieve_subtree.call_args == call(id=1, depth=1)
        assert cognite_client.assets.retrieve_subtree.call_count == 1


class TestAssetList:
    def test_get_events(self, cognite_client):
        cognite_client.events.list = mock.MagicMock()
        a = AssetList(resources=[Asset(id=1)], cognite_client=cognite_client)
        a.events()
        assert cognite_client.events.list.call_args == call(asset_ids=[1], limit=-1)
        assert cognite_client.events.list.call_count == 1

    def test_get_time_series(self, cognite_client):
        cognite_client.time_series.list = mock.MagicMock()
        a = AssetList(resources=[Asset(id=1)], cognite_client=cognite_client)
        a.time_series()
        assert cognite_client.time_series.list.call_args == call(asset_ids=[1], limit=-1)
        assert cognite_client.time_series.list.call_count == 1

    def test_get_sequences(self, cognite_client):
        cognite_client.sequences.list = mock.MagicMock()
        a = AssetList(resources=[Asset(id=1)], cognite_client=cognite_client)
        a.sequences()
        assert cognite_client.sequences.list.call_args == call(asset_ids=[1], limit=-1)
        assert cognite_client.sequences.list.call_count == 1

    def test_get_files(self, cognite_client):
        cognite_client.files.list = mock.MagicMock()
        a = AssetList(resources=[Asset(id=1)], cognite_client=cognite_client)
        a.files()
        assert cognite_client.files.list.call_args == call(asset_ids=[1], limit=-1)
        assert cognite_client.files.list.call_count == 1

    @pytest.mark.parametrize(
        "resource_class, resource_list_class, method",
        [(FileMetadata, FileMetadataList, "files"), (Event, EventList, "events")],
    )
    def test_get_related_resources_should_not_return_duplicates(self, resource_class, resource_list_class, method):
        r1 = resource_class(id=1)
        r2 = resource_class(id=2)
        r3 = resource_class(id=3)
        resources_a1 = resource_list_class([r1])
        resources_a2 = resource_list_class([r2, r3])
        resources_a3 = resource_list_class([r2, r3])

        mock_cognite_client = mock.MagicMock()
        mock_method = getattr(mock_cognite_client, method)
        mock_method.list.side_effect = [resources_a1, resources_a2, resources_a3]
        mock_method._config = mock.Mock(max_workers=3)

        assets = AssetList([Asset(id=1), Asset(id=2), Asset(id=3)], cognite_client=mock_cognite_client)
        assets._retrieve_chunk_size = 1

        resources = getattr(assets, method)()
        expected = [r1, r2, r3]
        assert expected == resources

    @pytest.mark.dsl
    def test_to_pandas_nullable_int(self, cognite_client):
        import pandas as pd

        for camel_case in [False, True]:
            assert (
                pd.Int64Dtype()
                == AssetList([Asset(parent_id=123), Asset(parent_id=None)]).to_pandas(camel_case=camel_case).dtypes[0]
            )
